import abc
import argparse
import glob
import os
from time import sleep
from typing import List, Iterable

from wai.logging import LOGGING_WARNING

from kasperl.api import MetaFileReader, check_dir
from seppl.io import InfiniteReader
from seppl.placeholders import PlaceholderSupporter, placeholder_list

GLOB_NAME_PLACEHOLDER = "{NAME}"
""" The glob placeholder for identifying other input files. """

POLL_ACTION_NOTHING = "nothing"
POLL_ACTION_MOVE = "move"
POLL_ACTION_DELETE = "delete"
POLL_ACTIONS = [
    POLL_ACTION_NOTHING,
    POLL_ACTION_MOVE, 
    POLL_ACTION_DELETE,
]


class PollDir(MetaFileReader, InfiniteReader, PlaceholderSupporter, abc.ABC):

    def __init__(self, dir_in: str = None, dir_out: str = None, poll_wait: float = None, process_wait: float = None,
                 action: str = None, extensions: List[str] = None,
                 other_input_files: List[str] = None, max_files: int = None, base_reader: str = None,
                 logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the reader.

        :param dir_in: the directory to poll for new files
        :type dir_in: str
        :param dir_out: the directory to move the files to before presenting them to the base reader
        :type dir_out: str
        :param poll_wait: the seconds to wait between polls
        :type poll_wait: float
        :param process_wait: the seconds to wait before processing the files (e.g., to be fully written to disk)
        :type process_wait: float
        :param action: the action to apply to the input files
        :type action: str
        :param extensions: the list of extensions to poll the directory for
        :type extensions: list
        :param other_input_files: other files that need to be present, glob expression (use placeholder GLOB_NAME_PLACEHOLDER)
        :type other_input_files: list
        :param max_files: the maximum number of files to poll (<1 for no limit)
        :type max_files: int
        :param base_reader: the base reader to use (command-line)
        :type base_reader: str
        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(base_reader=base_reader, logger_name=logger_name, logging_level=logging_level)
        self.dir_in = dir_in
        self.dir_out = dir_out
        self.poll_wait = poll_wait
        self.process_wait = process_wait
        self.action = action
        self.extensions = extensions
        self.other_input_files = other_input_files
        self.max_files = max_files
        self._inputs = None
        self._current_input = None
        self._base_reader = None
        self._actual_dir_in = None
        self._actual_dir_out = None

    def name(self) -> str:
        """
        Returns the name of the handler, used as sub-command.

        :return: the name
        :rtype: str
        """
        return "poll-dir"

    def description(self) -> str:
        """
        Returns a description of the reader.

        :return: the description
        :rtype: str
        """
        return "Polls a directory for files and presents them to the base reader."

    def _create_argparser(self) -> argparse.ArgumentParser:
        """
        Creates an argument parser. Derived classes need to fill in the options.

        :return: the parser
        :rtype: argparse.ArgumentParser
        """
        parser = super()._create_argparser()
        parser.add_argument("-i", "--dir_in", type=str, help="The directory to poll; " + placeholder_list(obj=self), required=True)
        parser.add_argument("-o", "--dir_out", type=str, help="The directory to move the files to; " + placeholder_list(obj=self), required=False)
        parser.add_argument("-w", "--poll_wait", type=float, help="The poll interval in seconds", required=False, default=1.0)
        parser.add_argument("-W", "--process_wait", type=float, help="The number of seconds to wait before processing the polled files (e.g., waiting for files to be fully written)", required=False, default=0.0)
        parser.add_argument("-a", "--action", choices=POLL_ACTIONS, help="The action to apply to the input files; 'move' moves the files to --dir_out directory", required=False, default=POLL_ACTION_MOVE)
        parser.add_argument("-e", "--extensions", type=str, help="The extensions of the files to poll (incl. dot)", required=True, nargs="+")
        parser.add_argument("-O", "--other_input_files", type=str, help="The glob expression(s) for capturing other files apart from the input files; use " + GLOB_NAME_PLACEHOLDER + " in the glob expression for the current name", required=False, default=None, nargs="*")
        parser.add_argument("-m", "--max_files", type=int, help="The maximum number of files in a single poll; <1 for unlimited", required=False, default=-1)
        return parser

    def _apply_args(self, ns: argparse.Namespace):
        """
        Initializes the object with the arguments of the parsed namespace.

        :param ns: the parsed arguments
        :type ns: argparse.Namespace
        """
        super()._apply_args(ns)
        self.dir_in = ns.dir_in
        self.dir_out = ns.dir_out
        self.poll_wait = ns.poll_wait
        self.process_wait = ns.process_wait
        self.action = ns.action
        self.extensions = ns.extensions
        self.other_input_files = ns.other_input_files
        self.max_files = ns.max_files

    def initialize(self):
        """
        Initializes the processing, e.g., for opening files or databases.
        """

        super().initialize()

        if self.poll_wait is None:
            self.poll_wait = 1.0
        if self.process_wait is None:
            self.process_wait = 0.0
        if self.action is None:
            self.action = POLL_ACTION_MOVE
        if (self.extensions is None) or (len(self.extensions) == 0):
            raise Exception("No extensions defined for polling!")
        if self.max_files is None:
            self.max_files = -1

        # check dirs
        self._actual_dir_in = self.session.expand_placeholders(self.dir_in)
        check_dir(self._actual_dir_in, "Input")
        if self.action == POLL_ACTION_MOVE:
            self._actual_dir_out = self.session.expand_placeholders(self.dir_out)
            check_dir(self._actual_dir_out, "Output")

    def list_files(self):
        """
        Generates the list of files.

        :return: list of files
        :rtype: List[str]
        """

        file_list = []
        self.logger().debug("Start listing files: %s" % self._actual_dir_in)

        try:
            for file_name in os.listdir(self._actual_dir_in):
                file_path = os.path.join(self._actual_dir_in, file_name)

                if os.path.isdir(file_path):
                    continue

                # monitored extension?
                if self.extensions is not None:
                    ext_lower = os.path.splitext(file_name)[1]
                    if ext_lower not in self.extensions:
                        self.logger().debug("%s does not match extensions: %s" % (file_name, str(self.extensions)))
                        continue

                file_list.append(file_path)

                # reached limit for poll?
                if self.max_files > 0:
                    if len(file_list) == self.max_files:
                        self.logger().debug("Reached maximum of %d files" % self.max_files)
                        break

            self.logger().debug("Finished listing files")

        except KeyboardInterrupt:
            return
        except:
            self.logger().exception("Failed listing files!")

        return file_list

    def read(self) -> Iterable:
        """
        Loads the data and returns the items one by one.

        :return: the data
        :rtype: Iterable
        """
        self.logger().info("Waiting for %s seconds before polling" % str(self.poll_wait))
        sleep(self.poll_wait)
        files = self.list_files()
        result = []
        if len(files) > 0:
            if self.process_wait > 0:
                self.logger().info("Waiting for %s seconds before processing" % str(self.process_wait))
                sleep(self.process_wait)
            result = self._read_files(files)

        # action?
        for file_path in files:
            if self.action == POLL_ACTION_DELETE:
                self.logger().debug("Deleting input: %s" % file_path)
                os.remove(file_path)
            elif self.action == POLL_ACTION_MOVE:
                self.logger().debug("Moving input: %s -> %s" % (file_path, self._actual_dir_out))
                os.rename(file_path, os.path.join(self._actual_dir_out, os.path.basename(file_path)))
            elif self.action == POLL_ACTION_NOTHING:
                pass
            else:
                raise Exception("Unhandled action: %s" % self.action)

            # other input files?
            if self.other_input_files is not None:
                for other_input_file in self.other_input_files:
                    other_files = glob.glob(os.path.join(self._actual_dir_in, other_input_file.replace(GLOB_NAME_PLACEHOLDER, os.path.splitext(file_path)[0])))
                    for other_file in other_files:
                        other_path = os.path.join(self._actual_dir_in, other_file)
                        if self.action == POLL_ACTION_DELETE:
                            self.logger().debug("Deleting other input: %s" % other_path)
                            os.remove(other_path)
                        elif self.action == POLL_ACTION_MOVE:
                            self.logger().debug("Moving other input: %s -> %s" % (other_path, self._actual_dir_out))
                            os.rename(other_path, os.path.join(self._actual_dir_out, os.path.basename(other_path)))
                        elif self.action == POLL_ACTION_NOTHING:
                            pass
                        else:
                            raise Exception("Unhandled action: %s" % self.action)

        for item in result:
            yield item

    def has_finished(self) -> bool:
        """
        Returns whether reading has finished.

        :return: True if finished
        :rtype: bool
        """
        return False



from typing import List, Dict


def list_classes() -> Dict[str, List[str]]:
    return {
        "seppl.io.Reader": [
            "kasperl.reader",
        ],
        "seppl.io.Filter": [
            "kasperl.filter",
        ],
        "seppl.io.Writer": [
            "kasperl.writer",
        ],
        "kasperl.api.Generator": [
            "kasperl.generator",
        ],
    }



import argparse
import os
import re
from typing import List

from seppl import AnyData
from seppl.io import BatchFilter
from wai.logging import LOGGING_WARNING

from kasperl.api import make_list, flatten_list, NameSupporter


class DiscardByName(BatchFilter):
    """
    Discards files based on list of image names and/or regular expressions that image names must match.
    """

    def __init__(self, names: List[str] = None, names_file: str = None, paths: List[str] = None,
                 regexps: List[str] = None, regexps_file: str = None,
                 remove_ext: bool = None, invert: bool = None,
                 logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the filter.

        :param names: the list of image names to drop
        :type names: list
        :param names_file: the text file with the image names to drop (one per line)
        :type names_file: str
        :param paths: the list of paths with files to ignore, ignored if None
        :type paths: list
        :param regexps: the regular expressions for dropping image names
        :type regexps: list
        :param regexps_file: the text file with the regexps for dropping image names (one per line)
        :type regexps_file: str
        :param remove_ext: whether to remove the extension before determining matches
        :type remove_ext: bool
        :param invert: whether to invert the matching sense
        :type invert: bool
        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(logger_name=logger_name, logging_level=logging_level)
        self.names = names
        self.names_file = names_file
        self.paths = paths
        self.regexps = regexps
        self.regexps_file = regexps_file
        self.remove_ext = remove_ext
        self.invert = invert
        self._names = None
        self._regexps = None

    def name(self) -> str:
        """
        Returns the name of the handler, used as sub-command.

        :return: the name
        :rtype: str
        """
        return "discard-by-name"

    def description(self) -> str:
        """
        Returns a description of the filter.

        :return: the description
        :rtype: str
        """
        return "Discards files based on list of names, list of paths and/or regular expressions that names must match."

    def accepts(self) -> List:
        """
        Returns the list of classes that are accepted.

        :return: the list of classes
        :rtype: list
        """
        return [AnyData]

    def generates(self) -> List:
        """
        Returns the list of classes that get produced.

        :return: the list of classes
        :rtype: list
        """
        return [AnyData]

    def _create_argparser(self) -> argparse.ArgumentParser:
        """
        Creates an argument parser. Derived classes need to fill in the options.

        :return: the parser
        :rtype: argparse.ArgumentParser
        """
        parser = super()._create_argparser()
        parser.add_argument("-i", "--names", type=str, help="The name(s) to drop.", required=False, nargs="*")
        parser.add_argument("-I", "--names_file", type=str, help="The text file with the name(s) to drop.", required=False, default=None)
        parser.add_argument("-p", "--paths", type=str, help="The directories with files to ignore.", required=False, nargs="*")
        parser.add_argument("-r", "--regexps", type=str, help="The regular expressions for matching name(s) to drop.", required=False, nargs="*")
        parser.add_argument("-R", "--regexps_file", type=str, help="The text file with regular expressions for matching name(s) to drop.", required=False, default=None)
        parser.add_argument("-e", "--remove_ext", action="store_true", help="Whether to remove the extension (and dot) before matching.")
        parser.add_argument("-V", "--invert", action="store_true", help="Whether to invert the matching sense.")
        return parser

    def _apply_args(self, ns: argparse.Namespace):
        """
        Initializes the object with the arguments of the parsed namespace.

        :param ns: the parsed arguments
        :type ns: argparse.Namespace
        """
        super()._apply_args(ns)
        self.names = ns.names
        self.names_file = ns.names_file
        self.paths = ns.paths
        self.regexps = ns.regexps
        self.regexps_file = ns.regexps_file
        self.remove_ext = ns.remove_ext
        self.invert = ns.invert

    def initialize(self):
        """
        Initializes the processing, e.g., for opening files or databases.
        """
        super().initialize()

        if self.remove_ext is None:
            self.remove_ext = False
        if self.invert is None:
            self.invert = False

        # names
        self._names = set()
        if self.names is not None:
            self._names.update(self.names)
        if self.names_file is not None:
            with open(self.names_file) as fp:
                lines = fp.readlines()
                for line in lines:
                    line = line.strip()
                    if len(line) > 0:
                        self._names.add(line)

        # paths
        if self.paths is not None:
            for path in self.paths:
                for f in os.listdir(path):
                    full = os.path.join(path, f)
                    if not os.path.isfile(full):
                        continue
                    if self.remove_ext:
                        f = os.path.splitext(f)[0]
                    self._names.add(f)

        self.logger().info("# names: %d" % len(self._names))

        # regexps
        self._regexps = list()
        if self.regexps is not None:
            for regexp in self.regexps:
                self._regexps.append(re.compile(regexp))
        if self.regexps_file is not None:
            with open(self.regexps_file) as fp:
                lines = fp.readlines()
                for line in lines:
                    line = line.strip()
                    if len(line) > 0:
                        self._regexps.append(re.compile(line))
        self.logger().info("# regexps: %d" % len(self._regexps))

    def _do_process(self, data):
        """
        Processes the data record(s).

        :param data: the record(s) to process
        :return: the potentially updated record(s)
        """
        # nothing to do?
        if (len(self._names) == 0) and (len(self._regexps) == 0):
            return data

        result = []
        for item in make_list(data):
            if not isinstance(item, NameSupporter):
                self.logger().warning("Does not implement '%s', cannot retrieve name for check, not discarding!" % str(NameSupporter))
                result.append(item)
                continue

            image_name = item.get_name()
            if self.remove_ext:
                image_name = os.path.splitext(image_name)[0]

            add = True

            # check against names
            if len(self._names) > 0:
                if image_name in self._names:
                    if not self.invert:
                        self.logger().info("Skipping based on name match: %s" % item.get_name())
                        add = False
                else:
                    if self.invert:
                        self.logger().info("Skipping based on no name match (invert): %s" % item.get_name())
                        add = False

            # check against regexps
            if add:
                for regexp in self._regexps:
                    if regexp.fullmatch(image_name) is not None:
                        if not self.invert:
                            self.logger().info("Skipping based on regexp match: %s" % item.get_name())
                            add = False
                            break
                    else:
                        if self.invert:
                            self.logger().info("Skipping based on no regexp match (invert): %s" % item.get_name())
                            add = False
                            break

            if add:
                result.append(item)

        return flatten_list(result)



import argparse
from typing import List

from wai.logging import LOGGING_WARNING

from seppl import AnyData
from seppl.io import BatchFilter
from kasperl.api import make_list, flatten_list, AnnotationHandler


class AnnotationsFromStorage(BatchFilter):
    """
    Retrieves the annotations from storage using the specified name.
    """

    def __init__(self, storage_name: str = None, logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the filter.

        :param storage_name: the name in internal storage to get the annotations from
        :type storage_name: str
        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(logger_name=logger_name, logging_level=logging_level)
        self.storage_name = storage_name

    def name(self) -> str:
        """
        Returns the name of the handler, used as sub-command.

        :return: the name
        :rtype: str
        """
        return "annotations-from-storage"

    def description(self) -> str:
        """
        Returns a description of the filter.

        :return: the description
        :rtype: str
        """
        return "Retrieves the annotations from storage using the specified name."

    def accepts(self) -> List:
        """
        Returns the list of classes that are accepted.

        :return: the list of classes
        :rtype: list
        """
        return [AnyData]

    def generates(self) -> List:
        """
        Returns the list of classes that get produced.

        :return: the list of classes
        :rtype: list
        """
        return [AnyData]

    def _create_argparser(self) -> argparse.ArgumentParser:
        """
        Creates an argument parser. Derived classes need to fill in the options.

        :return: the parser
        :rtype: argparse.ArgumentParser
        """
        parser = super()._create_argparser()
        parser.add_argument("-s", "--storage_name", type=str, help="The name in internal storage to get the annotations from.", required=True)
        return parser

    def _apply_args(self, ns: argparse.Namespace):
        """
        Initializes the object with the arguments of the parsed namespace.

        :param ns: the parsed arguments
        :type ns: argparse.Namespace
        """
        super()._apply_args(ns)
        self.storage_name = ns.storage_name

    def _do_process(self, data):
        """
        Processes the data record(s).

        :param data: the record(s) to process
        :return: the potentially updated record(s)
        """
        if self.storage_name not in self.session.storage:
            self.logger().error("Storage item not present: %s" % self.storage_name)
            return data

        ann = self.session.storage[self.storage_name]

        result = []
        for item in make_list(data):
            if isinstance(item, AnnotationHandler):
                self.logger().info("Using annotations from storage: %s" % self.storage_name)
                item.set_annotation(ann)
            else:
                self.logger().warning("Does not manage annotations: %s" % str(type(item)))
            result.append(item)
        return flatten_list(result)



import logging
from typing import Dict, Any

import seppl


class Session(seppl.Session):
    """
    Session object shared among reader, filter(s), writer.
    """
    storage: Dict[str, Any] = dict()
    """ for storing runtime data used within the pipeline. """

    logger: logging.Logger = logging.getLogger("kasperl")
    """ the global logger, derived frameworks should override this with their own. """



import argparse
import logging
import os
import re
from typing import List, Optional

from wai.logging import add_logging_level, init_logging, set_logging_level

from seppl.io import Splitter

DESCRIPTION = "Tool for locating files in directories that match certain patterns and store them in files."


def find_files_parser(prog: str, description: Optional[str]) -> argparse.ArgumentParser:
    """
    Creates the parser for the find files tool.

    :param prog: the name of the executable
    :type prog: str
    :param description: the description of the tool, uses default if None
    :type description: str
    :return: the parser instance
    :rtype: argparse.ArgumentParser
    """
    if description is None:
        description = DESCRIPTION
    result = argparse.ArgumentParser(prog=prog, description=description, formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    result.add_argument("-i", "--input", metavar="DIR", help="The dir(s) to scan for files.", default=None, type=str, required=True, nargs="+")
    result.add_argument("-r", "--recursive", action="store_true", help="Whether to search the directories recursively", required=False)
    result.add_argument("-o", "--output", metavar="FILE", help="The file to store the located file names in", type=str, required=True)
    result.add_argument("-m", "--match", metavar="REGEXP", help="The regular expression that the (full) file names must match to be included", default=None, type=str, required=False, nargs="*")
    result.add_argument("-n", "--not-match", metavar="REGEXP", help="The regular expression that the (full) file names must match to be excluded", default=None, type=str, required=False, nargs="*")
    result.add_argument("--split_ratios", type=int, default=None, help="The split ratios to use for generating the splits (int; must sum up to 100)", nargs="*")
    result.add_argument("--split_names", type=str, default=None, help="The split names to use as filename suffixes for the generated splits (before .ext)", nargs="*")
    result.add_argument("--split_name_separator", type=str, default="-", help="The separator to use between file name and split name", required=False)
    add_logging_level(result)
    return result


def _find(path: str, recursive: bool, match: List[str], not_match: List[str], files: List[str],
          logger: logging.Logger = None):
    """
    Locates files in the specified input directory and adds the matches to the files list.

    :param path: the directory to scan
    :type path: str
    :param recursive: whether to scan the dir(s) recursively
    :type recursive: bool
    :param match: the list regexp that the full paths of the files must match in order to be included (None: includes all)
    :type match: list
    :param not_match: the regexp that the full paths of the files must match in order to be excluded (None: includes all)
    :type not_match: list
    :param files: the list to add the matching files to
    :type files: list
    :param logger: the optional logger instance to use
    :type logger: logging.Logger
    """
    if logger is not None:
        logger.info("Entering: %s" % path)
    for f in os.listdir(path):
        if (f == ".") or (f == ".."):
            continue

        full = os.path.join(path, f)

        # dir?
        if os.path.isdir(full):
            if recursive:
                _find(full, recursive, match, not_match, files)
                logger.info("Back to: %s" % path)
            else:
                continue

        # match?
        is_match = True
        if match is not None:
            for pattern in match:
                if re.search(pattern, full) is None:
                    is_match = False
                    break
        if is_match and (not_match is not None):
            for pattern in not_match:
                if re.search(pattern, full) is not None:
                    is_match = False
                    break
        if is_match:
            files.append(full)


def find_files(paths: List[str], output: str, recursive=False,
               match: List[str] = None, not_match: List[str] = None,
               split_ratios: List[int] = None, split_names: List[str] = None,
               split_name_separator: str = "-", logger: logging.Logger = None):
    """
    Finds the files in the dir(s) and stores the full paths in text file(s).

    :param paths: the dir(s) to scan for files
    :type paths: list
    :param output: the output file to store the files in; when splitting, the split names get used as suffix (before .ext)
    :type output: str
    :param recursive: whether to scan the dir(s) recursively
    :type recursive: bool
    :param match: the list regexp that the full paths of the files must match in order to be included (None: includes all)
    :type match: list
    :param not_match: the regexp that the full paths of the files must match in order to be excluded (None: includes all)
    :type not_match: list
    :param split_ratios: the list of (int) ratios to use for splitting (must sum up to 100)
    :type split_ratios: list
    :param split_names: the list of suffixes to use for the splits
    :type split_names: list
    :param split_name_separator: the separator to use between file name and split name
    :type split_name_separator: str
    :param logger: the optional logger instance to use
    :type logger: logging.Logger
    """
    # locate
    all_files = []
    for d in paths:
        _find(d, recursive, match, not_match, all_files)
    if logger is not None:
        logger.info("# files found: %d" % len(all_files))

    # split?
    if (split_ratios is not None) or (split_names is not None):
        splitter = Splitter(split_ratios, split_names)
        splitter.initialize()
        # generate output names
        split_output = dict()
        parts = os.path.splitext(output)
        for name in split_names:
            split_output[name] = parts[0] + split_name_separator + name + parts[1]
        # split the files
        split_files = dict()
        for name in split_names:
            split_files[name] = []
        for f in all_files:
            split_files[splitter.next()].append(f)
        for name in split_names:
            if logger is not None:
                logger.info("# files in split '%s': %d" % (name, len(split_files[name])))
        # write files
        for name in split_names:
            if logger is not None:
                logger.info("Writing files to: %s" % split_output[name])
            with open(split_output[name], "w") as fp:
                for f in split_files[name]:
                    fp.write(f)
                    fp.write("\n")
    else:
        if logger is not None:
            logger.info("Writing files to: %s" % output)
        with open(output, "w") as fp:
            for f in all_files:
                fp.write(f)
                fp.write("\n")


def perform_find_files(env_var: Optional[str], args: List[str], prog: str, description: Optional[str], logger: logging.Logger):
    """
    The main method for parsing command-line arguments.

    :param env_var: the environment variable to obtain the logging level from, can be None
    :type env_var: str
    :param args: the commandline arguments, uses sys.argv if not supplied
    :type args: list
    :param prog: the name of the executable
    :type prog: str
    :param description: the description of the tool
    :type description: str
    :param logger: the logger instance to configure and use
    :type logger: logging.Logger
    """
    init_logging(env_var=env_var)
    parser = find_files_parser(prog, description)
    parsed = parser.parse_args(args=args)
    set_logging_level(logger, parsed.logging_level)
    find_files(parsed.input, parsed.output, recursive=parsed.recursive,
               match=parsed.match, not_match=parsed.not_match,
               split_ratios=parsed.split_ratios, split_names=parsed.split_names,
               split_name_separator=parsed.split_name_separator, logger=logger)



from ._annotations_from_storage import AnnotationsFromStorage
from ._annotations_to_storage import AnnotationsToStorage
from ._attach_metadata import AttachMetaData
from ._block import Block
from ._check_duplicate_filenames import CheckDuplicateFilenames, DUPLICATE_ACTIONS, DUPLICATE_ACTION_IGNORE, DUPLICATE_ACTION_WARN, DUPLICATE_ACTION_DROP, DUPLICATE_ACTION_ERROR
from ._copy_files import CopyFiles
from ._discard_by_name import DiscardByName
from ._list_to_sequence import ListToSequence
from ._max_records import MaxRecords
from ._metadata import MetaData
from ._metadata_from_name import MetaDataFromName
from ._metadata_to_placeholder import MetaDataToPlaceholder
from ._move_files import MoveFiles
from ._passthrough import PassThrough
from ._randomize_records import RandomizeRecords
from ._record_window import RecordWindow
from ._rename import Rename, RENAME_PLACEHOLDERS, RENAME_PH_NAME, RENAME_PH_COUNT, RENAME_PH_PDIR, RENAME_PH_SAME, RENAME_PH_PDIR_SUFFIX, RENAME_PH_EXT, RENAME_PH_OCCURRENCES, RENAME_PH_HELP
from ._sample import Sample
from ._set_metadata import SetMetaData
from ._set_placeholder import SetPlaceholder
from ._split_records import SplitRecords
from ._stop import Stop
from ._storage import StorageUpdater, DeleteStorage
from ._sub_process import SubProcess
from ._tee import Tee
from ._trigger import Trigger



from typing import List

from seppl import AnyData
from seppl.io import StreamWriter


class ConsoleWriter(StreamWriter):

    def name(self) -> str:
        """
        Returns the name of the handler, used as sub-command.

        :return: the name
        :rtype: str
        """
        return "console"

    def description(self) -> str:
        """
        Returns a description of the handler.

        :return: the description
        :rtype: str
        """
        return "Just prints the data to stdout."

    def accepts(self) -> List:
        """
        Returns the list of classes that are accepted.

        :return: the list of classes
        :rtype: list
        """
        return [AnyData]

    def write_stream(self, data):
        """
        Saves the data one by one.

        :param data: the data to write (single record or iterable of records)
        """
        print(data)



import argparse
from typing import Optional, List, Dict

from wai.logging import LOGGING_WARNING

from kasperl.api import SingleVariableGenerator, check_file


class TextFileGenerator(SingleVariableGenerator):
    """
    Outputs the lines from the text file (non-empty and not starting with #).
    """

    def __init__(self, text_file: str = None, var_name: str = None,
                 logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the generator.

        :param text_file: the text file with the values to use
        :type text_file: str
        :param var_name: the variable name
        :type var_name: str
        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(var_name=var_name, logger_name=logger_name, logging_level=logging_level)
        self.text_file = text_file

    def name(self) -> str:
        """
        Returns the name of the handler, used as sub-command.

        :return: the name
        :rtype: str
        """
        return "text-file"

    def description(self) -> str:
        """
        Returns a description of the handler.

        :return: the description
        :rtype: str
        """
        return "Outputs the lines from the text file (non-empty and not starting with #)."

    def _default_var_name(self) -> str:
        """
        Returns the default variable name.

        :return: the default name
        :rtype: str
        """
        return "v"

    def _create_argparser(self) -> argparse.ArgumentParser:
        """
        Creates an argument parser. Derived classes need to fill in the options.

        :return: the parser
        :rtype: argparse.ArgumentParser
        """
        parser = super()._create_argparser()
        parser.add_argument("-f", "--text_file", type=str, metavar="FILE", default=None, help="The text file with the values to use.", required=True)
        return parser

    def _apply_args(self, ns: argparse.Namespace):
        """
        Initializes the object with the arguments of the parsed namespace.

        :param ns: the parsed arguments
        :type ns: argparse.Namespace
        """
        super()._apply_args(ns)
        self.text_file = ns.text_file

    def _check(self) -> Optional[str]:
        """
        Hook method for performing checks.

        :return: None if checks successful, otherwise error message
        :rtype: str
        """
        result = super()._check()

        if result is None:
            result = check_file(self.text_file, "Text", raise_exc=False)

        return result

    def _do_generate(self) -> List[Dict[str, str]]:
        """
        Generates the variables.

        :return: the list of variable dictionaries
        :rtype: list
        """
        result = []
        with open(self.text_file, "r") as fp:
            values = fp.readlines()
            for value in values:
                value = value.strip()
                if (len(value) == 0) or value.startswith("#"):
                    continue
                result.append({self.var_name: str(value)})
        return result



from ._get_email import GetEmail, MESSAGE_ACTIONS, MESSAGE_ACTION_NONE, MESSAGE_ACTION_READ, MESSAGE_ACTION_DELETE
from ._list_files import ListFiles
from ._poll_dir import PollDir, POLL_ACTIONS, POLL_ACTION_NOTHING, POLL_ACTION_MOVE, POLL_ACTION_DELETE
from ._start import Start
from ._storage import StorageReader
from ._text_file import TextFileReader
from ._watch_dir import WatchDir, EVENTS, EVENT_MODIFIED, EVENT_CREATED, WATCH_ACTIONS, WATCH_ACTION_NOTHING, WATCH_ACTION_MOVE, WATCH_ACTION_DELETE, POLLING_TYPES, POLLING_TYPE_NEVER, POLLING_TYPE_INITIAL, POLLING_TYPE_ALWAYS



import argparse
from random import Random
from typing import List

from seppl import AnyData
from seppl.io import BatchFilter
from wai.logging import LOGGING_WARNING


class RandomizeRecords(BatchFilter):
    """
    Batch filter that randomizes the order of the records.
    """

    def __init__(self, seed: int = 1, logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the filter.

        :param seed: the seed value to use
        :type seed: int
        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(logger_name=logger_name, logging_level=logging_level)
        self.seed = seed
        self._rnd = None

    def name(self) -> str:
        """
        Returns the name of the handler, used as sub-command.

        :return: the name
        :rtype: str
        """
        return "randomize-records"

    def description(self) -> str:
        """
        Returns a description of the filter.

        :return: the description
        :rtype: str
        """
        return "Batch filter that randomizes the order of the records."

    def accepts(self) -> List:
        """
        Returns the list of classes that are accepted.

        :return: the list of classes
        :rtype: list
        """
        return [AnyData]

    def generates(self) -> List:
        """
        Returns the list of classes that get produced.

        :return: the list of classes
        :rtype: list
        """
        return [AnyData]

    def _create_argparser(self) -> argparse.ArgumentParser:
        """
        Creates an argument parser. Derived classes need to fill in the options.

        :return: the parser
        :rtype: argparse.ArgumentParser
        """
        parser = super()._create_argparser()
        parser.add_argument("-s", "--seed", type=int, help="The seed value to use for the randomization. Without seed value the order will differ between runs.", default=None, required=False)
        return parser

    def _apply_args(self, ns: argparse.Namespace):
        """
        Initializes the object with the arguments of the parsed namespace.

        :param ns: the parsed arguments
        :type ns: argparse.Namespace
        """
        super()._apply_args(ns)
        self.seed = ns.seed

    def initialize(self):
        """
        Initializes the processing, e.g., for opening files or databases.
        """
        super().initialize()
        self._rnd = Random(self.seed)

    def _requires_list_input(self) -> bool:
        """
        Returns whether lists are expected as input for the _process method.

        :return: True if list inputs are expected by the filter
        :rtype: bool
        """
        return True

    def _do_process(self, data):
        """
        Processes the data record(s).

        :param data: the record(s) to process
        :return: the potentially updated record(s)
        """
        if isinstance(data, list):
            result = list(data)
            self.logger().info("Randomizing %d records" % len(result))
            self._rnd.shuffle(result)
            return result
        else:
            self.logger().warning("Can only randomize lists, but received: %s" % str(type(data)))
            return data



import argparse
from typing import Optional, List, Dict

from wai.logging import LOGGING_WARNING
from kasperl.api import SingleVariableGenerator


class RangeGenerator(SingleVariableGenerator):
    """
    Generates a range of integers.
    """

    def __init__(self, from_: int = None, to: int = None, step: int = None, var_name: str = None,
                 logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the generator.

        :param from_: the starting value
        :type from_: int
        :param to: the end value (excluded)
        :type to: int
        :param step: the step between values
        :type step: int
        :param var_name: the variable name
        :type var_name: str
        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(var_name=var_name, logger_name=logger_name, logging_level=logging_level)
        self.from_ = from_
        self.to = to
        self.step = step

    def name(self) -> str:
        """
        Returns the name of the handler, used as sub-command.

        :return: the name
        :rtype: str
        """
        return "range"

    def description(self) -> str:
        """
        Returns a description of the handler.

        :return: the description
        :rtype: str
        """
        return "Iterates over a range of values."

    def _default_var_name(self) -> str:
        """
        Returns the default variable name.

        :return: the default name
        :rtype: str
        """
        return "i"

    def _create_argparser(self) -> argparse.ArgumentParser:
        """
        Creates an argument parser. Derived classes need to fill in the options.

        :return: the parser
        :rtype: argparse.ArgumentParser
        """
        parser = super()._create_argparser()
        parser.add_argument("-f", "--from", type=int, metavar="START", default=None, help="The starting value", required=True)
        parser.add_argument("-t", "--to", type=int, metavar="END", default=None, help="The end value (excluded)", required=True)
        parser.add_argument("-s", "--step", type=int, metavar="STEP", default=1, help="The increment between values", required=False)
        return parser

    def _apply_args(self, ns: argparse.Namespace):
        """
        Initializes the object with the arguments of the parsed namespace.

        :param ns: the parsed arguments
        :type ns: argparse.Namespace
        """
        super()._apply_args(ns)
        self.from_ = getattr(ns, "from")
        self.to = ns.to
        self.step = ns.step

    def _check(self) -> Optional[str]:
        """
        Hook method for performing checks.

        :return: None if checks successful, otherwise error message
        :rtype: str
        """
        result = super()._check()

        if self.step is None:
            self.step = 1

        if result is None:
            if self.from_ is None:
                result = "No FROM value provided!"
            elif self.to is None:
                result = "No TO value provided!"
            elif self.step == 0:
                result = "STEP cannot be 0!"
            elif self.step > 0:
                if self.from_ >= self.to:
                    result = "FROM must be smaller than TO (f=%s, t=%s, s=%s)!" % (str(self.from_), str(self.to), str(self.step))
            else:
                if self.from_ <= self.to:
                    result = "FROM must be less than TO (f=%s, t=%s, s=%s)!" % (str(self.from_), str(self.to), str(self.step))

        return result

    def _do_generate(self) -> List[Dict[str, str]]:
        """
        Generates the variables.

        :return: the list of variable dictionaries
        :rtype: list
        """
        result = []
        for i in range(self.from_, self.to, self.step):
            result.append({self.var_name: str(i)})
        return result



import types
from typing import List, Any, Optional


def make_list(data, cls=list) -> List:
    """
    Wraps the data item in a list if not already a list. Generators get turned into lists automatically.

    :param data: the data item to wrap if necessary
    :param cls: the type of class to check for
    :return: the list object
    :rtype: list
    """
    if isinstance(data, types.GeneratorType):
        return list(data)
    if not isinstance(data, cls):
        data = [data]
    return data


def flatten_list(data: List):
    """
    If the list contains only a single item, then it returns that instead of a list.

    :param data: the list to check
    :type data: list
    :return: the list or single item
    """
    if len(data) == 1:
        return data[0]
    else:
        return data


class AnnotationHandler:
    """
    Mixin for classes that manage annotations.
    """

    def has_annotation(self) -> bool:
        """
        Checks whether annotations are present.

        :return: True if annotations present
        :rtype: bool
        """
        raise NotImplementedError()

    def get_annotation(self) -> Optional[Any]:
        """
        Returns the annotations.

        :return: the annotations
        """
        raise NotImplementedError()

    def set_annotation(self, ann: Optional[Any]):
        """
        Sets the annotations.

        :param ann: the annotations
        """
        raise NotImplementedError()


class NameSupporter:
    """
    Mixin for classes that manage a name.
    """

    def get_name(self) -> str:
        """
        Returns the name.

        :return: the name
        :rtype: str
        """
        raise NotImplementedError()

    def set_name(self, name: str):
        """
        Sets the new name.

        :param name: the new name
        :type name: str
        """
        raise NotImplementedError()


class SourceSupporter:
    """
    Mixin for classes that manage a source, e.g., a file name.
    """

    def get_source(self) -> str:
        """
        Returns the source.

        :return: the source
        :rtype: str
        """
        raise NotImplementedError()



from typing import List, Iterable

from wai.logging import LOGGING_WARNING
from seppl import AnyData
from kasperl.api import Reader


class Start(Reader):
    """
    Dummy reader, simply forwards an empty string to trigger executions.
    """

    def __init__(self, logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the handler.

        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(logger_name=logger_name, logging_level=logging_level)
        self._finished = None

    def name(self) -> str:
        """
        Returns the name of the handler, used as sub-command.

        :return: the name
        :rtype: str
        """
        return "start"

    def description(self) -> str:
        """
        Returns a description of the reader.

        :return: the description
        :rtype: str
        """
        return "Dummy reader, simply forwards an empty string to trigger executions."

    def generates(self) -> List:
        """
        Returns the list of classes that get produced.

        :return: the list of classes
        :rtype: list
        """
        return [AnyData]

    def initialize(self):
        """
        Initializes the processing, e.g., for opening files or databases.
        """
        super().initialize()
        self._finished = False

    def read(self) -> Iterable:
        """
        Loads the data and returns the items one by one.

        :return: the data
        :rtype: Iterable
        """
        self._finished = True
        self.logger().info("Executing...")
        yield "start"

    def has_finished(self) -> bool:
        """
        Returns whether reading has finished.

        :return: True if finished
        :rtype: bool
        """
        return self._finished



import argparse
import os
from typing import List

from seppl import AnyData
from seppl.io import BatchFilter
from wai.logging import LOGGING_WARNING
from kasperl.api import make_list, flatten_list, NameSupporter, SourceSupporter

RENAME_PH_NAME = "{name}"
RENAME_PH_NAME_LOWERCASE = "{lname}"
RENAME_PH_NAME_UPPERCASE = "{uname}"
RENAME_PH_EXT = "{ext}"
RENAME_PH_EXT_LOWERCASE = "{lext}"
RENAME_PH_EXT_UPPERCASE = "{uext}"
RENAME_PH_OCCURRENCES = "{occurrences}"
RENAME_PH_COUNT = "{count}"
RENAME_PH_PDIR = "{[p]+dir}"
RENAME_PH_PDIR_SUFFIX = "pdir}"

RENAME_PLACEHOLDERS = [
    RENAME_PH_NAME,
    RENAME_PH_NAME_LOWERCASE,
    RENAME_PH_NAME_UPPERCASE,
    RENAME_PH_EXT,
    RENAME_PH_EXT_LOWERCASE,
    RENAME_PH_EXT_UPPERCASE,
    RENAME_PH_OCCURRENCES,
    RENAME_PH_COUNT,
    RENAME_PH_PDIR,
]

RENAME_PH_HELP = {
    RENAME_PH_NAME: "the name of the file, without path or extension.",
    RENAME_PH_NAME_LOWERCASE: "the lower-case name of the file, without path or extension.",
    RENAME_PH_NAME_UPPERCASE: "the upper-case name of the file, without path or extension.",
    RENAME_PH_EXT: "the extension of the file (incl dot).",
    RENAME_PH_EXT_LOWERCASE: "the lower-case extension of the file (incl dot).",
    RENAME_PH_EXT_UPPERCASE: "the upper-case extension of the file (incl dot).",
    RENAME_PH_OCCURRENCES: "the number of times this name (excl extension) has been encountered.",
    RENAME_PH_COUNT: "the number of files encountered so far.",
    RENAME_PH_PDIR: "the parent directory of the file: 'p': immediate parent, the more the p's the higher up in the hierarchy.",
}

RENAME_PH_SAME = RENAME_PH_NAME + RENAME_PH_EXT


def _format_help() -> str:
    """
    Generates the help for the name format.

    :return: the generated help
    """
    result = "Available placeholders:\n"
    for ph in RENAME_PLACEHOLDERS:
        result += "- %s: %s\n" % (ph, RENAME_PH_HELP[ph])

    return result


class Rename(BatchFilter):
    """
    Renames files using a user-supplied format.
    """

    def __init__(self, name_format: str = RENAME_PH_SAME, logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the filter.

        :param name_format: the format to use for the file name
        :type name_format: str
        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(logger_name=logger_name, logging_level=logging_level)
        self.name_format = name_format
        self._count = None
        self._occurrences = None

    def name(self) -> str:
        """
        Returns the name of the handler, used as sub-command.

        :return: the name
        :rtype: str
        """
        return "rename"

    def description(self) -> str:
        """
        Returns a description of the filter.

        :return: the description
        :rtype: str
        """
        return "Renames files using a user-supplied format."

    def accepts(self) -> List:
        """
        Returns the list of classes that are accepted.

        :return: the list of classes
        :rtype: list
        """
        return [AnyData]

    def generates(self) -> List:
        """
        Returns the list of classes that get produced.

        :return: the list of classes
        :rtype: list
        """
        return [AnyData]

    def _create_argparser(self) -> argparse.ArgumentParser:
        """
        Creates an argument parser. Derived classes need to fill in the options.

        :return: the parser
        :rtype: argparse.ArgumentParser
        """
        parser = super()._create_argparser()
        parser.add_argument("-f", "--name-format", type=str, help="The format for the new name.\n" + _format_help(), default=RENAME_PH_SAME, required=False)
        return parser

    def _apply_args(self, ns: argparse.Namespace):
        """
        Initializes the object with the arguments of the parsed namespace.

        :param ns: the parsed arguments
        :type ns: argparse.Namespace
        """
        super()._apply_args(ns)
        self.name_format = ns.name_format

    def initialize(self):
        """
        Initializes the processing, e.g., for opening files or databases.
        """
        super().initialize()
        self._count = 0
        self._occurrences = dict()

    def _parent_dir(self, path: str, pattern: str) -> str:
        """
        Returns the parent directory that corresponds to the pdir pattern.

        :param path: the directory path of the file
        :param pattern: the pdir pattern (without curly brackets)
        :return: the parent dir, empty string if failed to determine or too high
        """
        pattern = pattern.replace("dir", "")
        num = pattern.count("p")
        if num == 0:
            return ""
        orig_num = num
        orig_path = path
        pdir = os.path.basename(path)
        while (num > 1) and (len(path) > 0):
            path = os.path.dirname(path)
            pdir = os.path.basename(path)
            num -= 1
        if pdir == "":
            self.logger().warning("Number of parents (%d) is too high for path (%s) to generate a name!" % (orig_num, orig_path))
        return pdir

    def _do_process(self, data):
        """
        Processes the data record(s).

        :param data: the record(s) to process
        :return: the potentially updated record(s)
        """
        # nothing to do?
        if self.name_format == RENAME_PH_SAME:
            return data

        result = []
        for item in make_list(data):
            # name
            if not isinstance(item, NameSupporter):
                self.logger().warning("Does not implement '%s', cannot retrieve name for check, not discarding!" % str(NameSupporter))
                result.append(item)
                continue
            name, ext = os.path.splitext(item.get_name())

            # source
            if not isinstance(item, SourceSupporter):
                self.logger().warning("Does not implement '%s', cannot retrieve name for check, not discarding!" % str(SourceSupporter))
                result.append(item)
                continue
            if item.get_source() is None:
                path = "."
            else:
                path = os.path.dirname(item.get_source())

            # increment counter
            self._count += 1

            # occurrences
            if name not in self._occurrences:
                self._occurrences[name] = 1
            else:
                self._occurrences[name] += 1

            # generate new name
            name_new = self.name_format
            name_new = name_new.replace(RENAME_PH_NAME, name)
            name_new = name_new.replace(RENAME_PH_NAME_LOWERCASE, name.lower())
            name_new = name_new.replace(RENAME_PH_NAME_UPPERCASE, name.upper())
            name_new = name_new.replace(RENAME_PH_EXT, ext)
            name_new = name_new.replace(RENAME_PH_EXT_LOWERCASE, ext.lower())
            name_new = name_new.replace(RENAME_PH_EXT_UPPERCASE, ext.upper())
            name_new = name_new.replace(RENAME_PH_COUNT, str(self._count))
            name_new = name_new.replace(RENAME_PH_OCCURRENCES, str(self._occurrences[name]))
            while RENAME_PH_PDIR_SUFFIX in name_new:
                index = name_new.index(RENAME_PH_PDIR_SUFFIX)
                start = name_new.rfind("{", 0, index)
                if start == -1:
                    self.logger().error("Found '%s' without starting '{' in name format!" % RENAME_PH_PDIR_SUFFIX)
                    break
                pattern = name_new[start + 1:index + 4]
                pdir = self._parent_dir(path, pattern)
                self.logger().info("Parent dir: %s -> %s" % (pattern, pdir))
                name_new = name_new.replace("{%s}" % pattern, pdir)

            self.logger().info("Result: %s -> %s" % (name, name_new))

            item_new = item.duplicate(source=os.path.join(path, name_new), name=name_new)
            result.append(item_new)

        return flatten_list(result)



import abc
import argparse
from typing import List, Dict

from wai.logging import LOGGING_WARNING
from seppl import Initializable, Plugin
import seppl.io


class BatchWriter(seppl.io.BatchWriter, Initializable, abc.ABC):
    """
    Ancestor for dataset batch writers.
    """

    def __init__(self, logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the reader.

        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(logger_name=logger_name, logging_level=logging_level)


class SplittableBatchWriter(BatchWriter, abc.ABC):
    """
    Ancestor for dataset batch writers.
    """

    def __init__(self, split_names: List[str] = None, split_ratios: List[int] = None, split_group: str = None,
                 logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the reader.

        :param split_names: the names of the splits, no splitting if None
        :type split_names: list
        :param split_ratios: the integer ratios of the splits (must sum up to 100)
        :type split_ratios: list
        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(logger_name=logger_name, logging_level=logging_level)
        self.split_names = None
        self.split_ratios = None
        self.split_group = None
        self.splitter = None
        seppl.io.init_splitting_params(self, split_names=split_names, split_ratios=split_ratios)

    def _create_argparser(self) -> argparse.ArgumentParser:
        """
        Creates an argument parser. Derived classes need to fill in the options.

        :return: the parser
        :rtype: argparse.ArgumentParser
        """
        parser = super()._create_argparser()
        seppl.io.add_splitting_params(parser)
        return parser

    def _apply_args(self, ns: argparse.Namespace):
        """
        Initializes the object with the arguments of the parsed namespace.

        :param ns: the parsed arguments
        :type ns: argparse.Namespace
        """
        super()._apply_args(ns)
        seppl.io.transfer_splitting_params(ns, self)

    def initialize(self):
        """
        Initializes the processing, e.g., for opening files or databases.
        """
        super().initialize()
        seppl.io.initialize_splitting(self)


class StreamWriter(seppl.io.StreamWriter, Initializable, abc.ABC):
    """
    Ancestor for dataset stream writers.
    """

    def __init__(self, logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the reader.

        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(logger_name=logger_name, logging_level=logging_level)


class SplittableStreamWriter(StreamWriter, abc.ABC):
    """
    Ancestor for dataset stream writers.
    """

    def __init__(self, split_names: List[str] = None, split_ratios: List[int] = None, split_group: str = None,
                 logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the reader.

        :param split_names: the names of the splits, no splitting if None
        :type split_names: list
        :param split_ratios: the integer ratios of the splits (must sum up to 100)
        :type split_ratios: list
        :param split_group: the regular expression with a single group used for keeping items in the same split, e.g., for identifying the base name of a file or the sample ID
        :type split_group: str
        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(logger_name=logger_name, logging_level=logging_level)
        # define the members
        self.split_names = None
        self.split_ratios = None
        self.split_group = None
        self.splitter = None
        # initialize the members
        seppl.io.init_splitting_params(self, split_names=split_names, split_ratios=split_ratios, split_group=split_group)

    def _create_argparser(self) -> argparse.ArgumentParser:
        """
        Creates an argument parser. Derived classes need to fill in the options.

        :return: the parser
        :rtype: argparse.ArgumentParser
        """
        parser = super()._create_argparser()
        seppl.io.add_splitting_params(parser)
        return parser

    def _apply_args(self, ns: argparse.Namespace):
        """
        Initializes the object with the arguments of the parsed namespace.

        :param ns: the parsed arguments
        :type ns: argparse.Namespace
        """
        super()._apply_args(ns)
        seppl.io.transfer_splitting_params(ns, self)

    def initialize(self):
        """
        Initializes the processing, e.g., for opening files or databases.
        """
        super().initialize()
        seppl.io.initialize_splitting(self)


def parse_writer(writer: str, available_writers: Dict[str, Plugin]) -> seppl.io.Writer:
    """
    Parses the command-line and instantiates the writer.

    :param writer: the command-line to parse
    :type writer: str
    :param available_writers: the writers to use for parsing
    :type available_writers: dict
    :return: the writer
    :rtype: seppl.io.Writer
    """
    from seppl import split_args, split_cmdline, args_to_objects

    if writer is None:
        raise Exception("No writer command-line supplied!")
    valid = dict()
    valid.update(available_writers)
    args = split_args(split_cmdline(writer), list(valid.keys()))
    objs = args_to_objects(args, valid, allow_global_options=False)
    if len(objs) == 1:
        if isinstance(objs[0], seppl.io.Writer):
            result = objs[0]
        else:
            raise Exception("Expected instance of Writer but got: %s" % str(type(objs[0])))
    else:
        raise Exception("Expected to obtain one writer from '%s' but got %d instead!" % (writer, len(objs)))
    return result


def add_annotations_only_writer_param(parser: argparse.ArgumentParser):
    """
    Adds the --annotations_only parameter to the parser, as used by writers of type AnnotationOnlyWriter.

    :param parser: the parser
    :type parser: argparse.ArgumentParser
    """
    parser.add_argument("--annotations_only", action="store_true", help="Outputs only the annotations.")


class AnnotationsOnlyWriter:
    """
    Mixin for writers that can output the annotations by itself.
    """
    pass



from ._comparison import compare_values, COMPARISONS, COMPARISON_HELP, COMPARISONS_EXT, COMPARISON_EXT_HELP, \
    COMPARISON_LESSTHAN, COMPARISON_LESSOREQUAL, COMPARISON_EQUAL, COMPARISON_NOTEQUAL, COMPARISON_GREATEROREQUAL, \
    COMPARISON_GREATERTHAN, COMPARISON_CONTAINS, COMPARISON_MATCHES
from ._session import Session
from ._data import make_list, flatten_list, NameSupporter, SourceSupporter, AnnotationHandler
from ._generator import Generator, SingleVariableGenerator, test_generator, perform_generator_test, compile_generator_vars_list
from ._exec import execute_pipeline, perform_pipeline_execution, load_pipeline, PIPELINE_FORMATS, PIPELINE_FORMAT_CMDLINE, PIPELINE_FORMAT_FILE
from ._find import find_files_parser, find_files, perform_find_files
from ._reader import Reader, parse_reader, AnnotationsOnlyReader, add_annotations_only_reader_param, MetaFileReader
from ._filter import BatchFilter, parse_filter
from ._writer import BatchWriter, SplittableBatchWriter, StreamWriter, SplittableStreamWriter, parse_writer, \
    AnnotationsOnlyWriter, add_annotations_only_writer_param
from ._utils import strip_suffix, locate_file, load_function, safe_deepcopy, annotation_to_name, check_dir, check_file
from ._conversion import parse_conversion_args, print_conversion_usage, perform_conversion
from ._help import CommandlineParameter, params_to_parser, param_to_parser, params_to_short, param_to_short, param_to_help
from ._plots import Plot, XYPlot, SequencePlot



import argparse
from typing import List

from wai.logging import LOGGING_WARNING

from seppl import AnyData, MetaDataHandler
from seppl.io import BatchFilter
from kasperl.api import make_list, flatten_list, compare_values, \
    COMPARISONS_EXT, COMPARISON_EQUAL, COMPARISON_CONTAINS, COMPARISON_MATCHES, COMPARISON_EXT_HELP


class Stop(BatchFilter):
    """
    Raises an exception when the expression evaluates to True, stopping the pipeline.
    """

    def __init__(self, field: str = None, comparison: str = COMPARISON_EQUAL, value=None,
                 stop_message: str = None, logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the filter.

        :param field: the name of the meta-data field to perform the comparison on
        :type field: str
        :param comparison: the comparison to perform
        :type comparison: str
        :param value: the value to compare with
        :param stop_message: the message to use in the exception
        :type stop_message: str
        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(logger_name=logger_name, logging_level=logging_level)
        self.field = field
        self.comparison = comparison
        self.value = value
        self.stop_message = stop_message

    def name(self) -> str:
        """
        Returns the name of the handler, used as sub-command.

        :return: the name
        :rtype: str
        """
        return "stop"

    def description(self) -> str:
        """
        Returns a description of the handler.

        :return: the description
        :rtype: str
        """
        return "Raises an exception when the expression evaluates to True, stopping the pipeline. " \
               "Never raises an exception if no meta-data field specified. " \
               "Performs the following comparison: METADATA_VALUE COMPARISON VALUE."

    def accepts(self) -> List:
        """
        Returns the list of classes that are accepted.

        :return: the list of classes
        :rtype: list
        """
        return [AnyData]

    def generates(self) -> List:
        """
        Returns the list of classes that get produced.

        :return: the list of classes
        :rtype: list
        """
        return [AnyData]

    def _create_argparser(self) -> argparse.ArgumentParser:
        """
        Creates an argument parser. Derived classes need to fill in the options.

        :return: the parser
        :rtype: argparse.ArgumentParser
        """
        parser = super()._create_argparser()
        parser.add_argument("--field", type=str, help="The meta-data field to use in the comparison", default=None, required=False)
        parser.add_argument("--comparison", choices=COMPARISONS_EXT, default=COMPARISON_EQUAL, help="How to compare the value with the meta-data value; " + COMPARISON_EXT_HELP
                            + "; in case of '" + COMPARISON_CONTAINS + "' and '" + COMPARISON_MATCHES + "' the supplied value represents the substring to find/regexp to search with", required=False)
        parser.add_argument("--value", type=str, help="The value to use in the comparison", default=None, required=False)
        parser.add_argument("--stop_message", type=str, help="The message to use in the exception.", default=None, required=False)
        return parser

    def _apply_args(self, ns: argparse.Namespace):
        """
        Initializes the object with the arguments of the parsed namespace.

        :param ns: the parsed arguments
        :type ns: argparse.Namespace
        """
        super()._apply_args(ns)
        self.field = ns.field
        self.comparison = ns.comparison
        self.value = ns.value
        self.stop_message = ns.stop_message

    def initialize(self):
        """
        Initializes the processing, e.g., for opening files or databases.
        """
        super().initialize()
        if (self.field is not None) and (self.value is None):
            raise Exception("No value provided to compare with!")

    def _do_process(self, data):
        """
        Processes the data record(s).

        :param data: the record(s) to process
        :return: the potentially updated record(s)
        """
        result = []
        for item in make_list(data):
            # evaluate expression?
            meta = None
            if self.field is not None:
                if isinstance(item, MetaDataHandler):
                    if item.has_metadata():
                        meta = item.get_metadata()
            if meta is not None:
                v1 = meta[self.field]
                v2 = self.value
                comp_result = compare_values(v1, self.comparison, v2)
                comp = str(meta[self.field]) + " " + self.comparison + " " + str(self.value) + " = " + str(comp_result)
                msg = "Field '%s': '%s'" % (self.field, comp)
                self.logger().info(msg)
                if comp_result:
                    self.session.stopped = True
                    if self.stop_message is None:
                        self.logger().error(msg)
                    else:
                        self.logger().error(self.stop_message)

            result.append(item)

        return flatten_list(result)



import abc
import argparse
import csv
import json
import os
from typing import List, Optional, Any, Dict

from wai.logging import LOGGING_WARNING

from kasperl.api import make_list, flatten_list
from seppl import MetaDataHandler, AnyData
from seppl.io import BatchFilter

METADATA_FORMAT_CSV = "csv"
METADATA_FORMAT_JSON = "json"
METADATA_FORMATS = [
    METADATA_FORMAT_CSV,
    METADATA_FORMAT_JSON,
]
METADATA_FORMAT_EXTENSIONS = {
    METADATA_FORMAT_CSV: ".csv",
    METADATA_FORMAT_JSON: ".json",
}


class AttachMetaData(BatchFilter, abc.ABC):
    """
    Attaches meta-data to the data passing through. Loads the data from the specified directory based on the data's name.
    """

    def __init__(self, metadata_dir: str = None, metadata_format: str = None,
                 logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the filter.

        :param metadata_dir: the directory with the meta-data files to load/attach
        :type metadata_dir: str
        :param metadata_format: the format that the meta-data is in
        :type metadata_format: str
        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(logger_name=logger_name, logging_level=logging_level)
        self.metadata_dir = metadata_dir
        self.metadata_format = metadata_format
        self.attached = 0
        self.missed = 0

    def name(self) -> str:
        """
        Returns the name of the handler, used as sub-command.

        :return: the name
        :rtype: str
        """
        return "attach-metadata"

    def description(self) -> str:
        """
        Returns a description of the filter.

        :return: the description
        :rtype: str
        """
        return "Attaches meta-data to the data passing through. " \
            + "Loads the data from the specified directory based on the data's name. " \
            + "In case of CSV, a header row is expected and the first column contains the keys and the second the values."

    def accepts(self) -> List:
        """
        Returns the list of classes that are accepted.

        :return: the list of classes
        :rtype: list
        """
        return [AnyData]

    def generates(self) -> List:
        """
        Returns the list of classes that get produced.

        :return: the list of classes
        :rtype: list
        """
        return [AnyData]

    def _create_argparser(self) -> argparse.ArgumentParser:
        """
        Creates an argument parser. Derived classes need to fill in the options.

        :return: the parser
        :rtype: argparse.ArgumentParser
        """
        parser = super()._create_argparser()
        parser.add_argument("-m", "--metadata_dir", type=str, help="The directory with the meta-data files to load/attach.", required=True)
        parser.add_argument("-f", "--metadata_format", choices=METADATA_FORMATS, default=METADATA_FORMAT_JSON, help="The format of the meta-data.", required=False)
        return parser

    def _apply_args(self, ns: argparse.Namespace):
        """
        Initializes the object with the arguments of the parsed namespace.

        :param ns: the parsed arguments
        :type ns: argparse.Namespace
        """
        super()._apply_args(ns)
        self.metadata_dir = ns.metadata_dir
        self.metadata_format = ns.metadata_format

    def initialize(self):
        """
        Initializes the processing, e.g., for opening files or databases.
        """
        super().initialize()
        if self.metadata_dir is None:
            raise Exception("No meta-data directory provided!")
        if self.metadata_format is None:
            self.metadata_format = METADATA_FORMAT_JSON
        if self.metadata_format not in METADATA_FORMATS:
            raise Exception("Unknown meta-data format: %s" % self.metadata_format)

    @abc.abstractmethod
    def _get_name(self, item) -> Optional[str]:
        """
        Returns the name of the item.

        :param item: the item to get the name for
        :return: the name or None if not available
        :rtype: str or None
        """
        raise NotImplementedError()

    def _load_metadata(self, name: str) -> Optional[Dict[str, Any]]:
        """
        Loads the meta-data if available.

        :param name: the name to use for locating the meta-data
        :type name: str
        :return: the meta-data, None if not found
        :rtype: dict or None
        """
        name = os.path.splitext(name)[0] + METADATA_FORMAT_EXTENSIONS[self.metadata_format]
        path = os.path.join(self.session.expand_placeholders(self.metadata_dir), name)
        if not os.path.exists(path):
            return None

        if self.metadata_format == METADATA_FORMAT_CSV:
            result = dict()
            with open(path, newline='') as fp:
                reader = csv.reader(fp)
                first = True
                for row in reader:
                    # ignore header
                    if first:
                        first = False
                        continue
                    result[str(row[0])] = row[1]
            return result

        elif self.metadata_format == METADATA_FORMAT_JSON:
            with open(path, "r") as fp:
                return dict(json.load(fp))

        else:
            raise Exception("Unhandled meta-data format: %s" % self.metadata_format)

    def _do_process(self, data):
        """
        Processes the data record(s).

        :param data: the record(s) to process
        :return: the potentially updated record(s)
        """
        result = []

        for item in make_list(data):
            # can we attach meta-data?
            if not isinstance(item, MetaDataHandler):
                self.logger().warning("Data does not manage meta-data, cannot attach meta-data: %s" % str(item))
                result.append(item)
                self.missed += 1
                continue

            # do we have a name?
            name = self._get_name(item)
            if name is None:
                self.logger().warning("Failed to determine name from data, cannot attach meta-data: %s" % str(item))
                result.append(item)
                self.missed += 1
                continue

            meta = self._load_metadata(name)
            if meta is None:
                self.logger().warning("No meta-data located for: %s" % name)
                self.missed += 1
            else:
                if item.has_metadata():
                    item.get_metadata().update(meta)
                else:
                    item.set_metadata(meta)
                self.attached += 1
            result.append(item)

        return flatten_list(result)

    def finalize(self):
        """
        Finishes the processing, e.g., for closing files or databases.
        """
        super().finalize()
        self.logger().info("# attached: %d" % self.attached)
        self.logger().info("# missed: %d" % self.missed)



import types
from typing import List

from kasperl.api import make_list
from seppl import AnyData
from seppl.io import StreamFilter


class ListToSequence(StreamFilter):
    """
    Forwards the individual items of the incoming list.
    """

    def name(self) -> str:
        """
        Returns the name of the handler, used as sub-command.

        :return: the name
        :rtype: str
        """
        return "list-to-sequence"

    def description(self) -> str:
        """
        Returns a description of the handler.

        :return: the description
        :rtype: str
        """
        return "Forwards the individual items of the incoming list."

    def accepts(self) -> List:
        """
        Returns the list of classes that are accepted.

        :return: the list of classes
        :rtype: list
        """
        return [AnyData]

    def generates(self) -> List:
        """
        Returns the list of classes that get produced.

        :return: the list of classes
        :rtype: list
        """
        return [AnyData]

    def _do_process_stream(self, data):
        """
        Filters the data.

        :param data: the data to filter
        """
        for item in make_list(data):
            if isinstance(item, list):
                self._stream_output.extend(item)
            elif isinstance(item, types.GeneratorType):
                for sub in item:
                    self._stream_output.append(sub)
            else:
                self._stream_output.append(item)

    def _do_process(self, data):
        """
        Processes the data record(s).

        :param data: the record(s) to process
        :return: the potentially updated record(s)
        """
        return data



import abc
import argparse
import os
import smtplib
from email import encoders
from email.mime.base import MIMEBase
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from typing import List, Union

from dotenv import load_dotenv
from wai.logging import LOGGING_WARNING

from kasperl.api import make_list, StreamWriter
from seppl import AnyData
from seppl.placeholders import InputBasedPlaceholderSupporter, placeholder_list

SMTP_HOST = "SMTP_HOST"
SMTP_PORT = "SMTP_PORT"
SMTP_STARTTLS = "SMTP_STARTTLS"
SMTP_USER = "SMTP_USER"
SMTP_PW = "SMTP_PW"
SMTP_ENVS = [
    SMTP_HOST,
    SMTP_PORT,
    SMTP_STARTTLS,
    SMTP_USER,
    SMTP_PW,
]


class SendEmail(StreamWriter, InputBasedPlaceholderSupporter, abc.ABC):

    def __init__(self, dotenv_path: str = None, email_from: str = None, email_to: Union[str, List[str]] = None,
                 subject: str = None, body: str = None, body_file: str = None,
                 logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the writer.

        :param dotenv_path: the path to the .env file to use for initializing the SMTP environment vars
        :type dotenv_path: str
        :param email_from: the FROM address to use in the email
        :type email_from: str
        :param email_to: the TO address(es) to use in the email
        :type email_to: str or list
        :param subject: the SUBJECT for the email
        :type subject: str
        :param body: the BODY for the email
        :type body: str
        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(logger_name=logger_name, logging_level=logging_level)
        self.dotenv_path = dotenv_path
        self.email_from = email_from
        self.email_to = email_to
        self.subject = subject
        self.body = body
        self.body_file = body_file
        self._body = None
        self._dotenv_loaded = False
        self._server = None

    def name(self) -> str:
        """
        Returns the name of the handler, used as sub-command.

        :return: the name
        :rtype: str
        """
        return "send-email"

    def description(self) -> str:
        """
        Returns a description of the writer.

        :return: the description
        :rtype: str
        """
        return "Attaches the incoming file(s) and sends them to the specified email address(es)."

    def _create_argparser(self) -> argparse.ArgumentParser:
        """
        Creates an argument parser. Derived classes need to fill in the options.

        :return: the parser
        :rtype: argparse.ArgumentParser
        """
        parser = super()._create_argparser()
        parser.add_argument("-d", "--dotenv_path", metavar="FILE", type=str, help="The .env file to load the SMTP environment variables form (" + "|".join(SMTP_ENVS) + "); tries to load .env from the current directory if not specified; " + placeholder_list(obj=self), required=False, default=None)
        parser.add_argument("-f", "--email_from", metavar="EMAIL", type=str, help="The email address to use for FROM; placeholders get automatically expanded.", default=None, required=True)
        parser.add_argument("-t", "--email_to", metavar="EMAIL", type=str, help="The email address(es) to send the email TO; placeholders get automatically expanded.", default=None, required=True, nargs="+")
        parser.add_argument("-s", "--subject", metavar="SUBJECT", type=str, help="The SUBJECT for the email; placeholders get automatically expanded.", default=None, required=False)
        parser.add_argument("-b", "--body", metavar="TEXT", type=str, help="The email body to use, splits on '\n' strings (not newlines!); placeholders get automatically expanded.", default=None, required=False)
        parser.add_argument("-B", "--body_file", metavar="FILE", type=str, help="The file with the email body to use; placeholders get automatically expanded.", default=None, required=False)
        return parser

    def _apply_args(self, ns: argparse.Namespace):
        """
        Initializes the object with the arguments of the parsed namespace.

        :param ns: the parsed arguments
        :type ns: argparse.Namespace
        """
        super()._apply_args(ns)
        self.dotenv_path = ns.dotenv_path
        self.email_from = ns.email_from
        self.email_to = ns.email_to
        self.subject = ns.subject
        self.body = ns.body
        self.body_file = ns.body_file

    def initialize(self):
        """
        Initializes the processing, e.g., for opening files or databases.
        """
        super().initialize()
        self._dotenv_loaded = False
        self._server = None
        if self.email_from is None:
            raise Exception("No FROM address specified!")
        if self.email_to is None:
            raise Exception("At least one TO address must be specified!")
        if self.subject is None:
            self.subject = ""
        self._body = None
        if self.body_file is not None:
            path = self.session.expand_placeholders(self.body_file)
            if not os.path.exists(path):
                raise Exception("File with email body does not exist: %s" % path)
            if os.path.isdir(path):
                raise Exception("File with email body points to a directory: %s" % path)
            self.logger().info("Loading email body file: %s" % path)
            with open(path, "r") as fp:
                self._body = [x.strip() for x in fp.readlines()]
        if (self._body is None) and (self.body is not None):
            self._body = self.body.split("\\n")
        if self._body is None:
            self._body = []

    def accepts(self) -> List:
        """
        Returns the list of classes that are accepted.

        :return: the list of classes
        :rtype: list
        """
        return [AnyData]

    def _attach_data(self, message: MIMEMultipart, data: bytes, name: str,
                     mime_main: str = "application", mime_sub: str = "octet-stream"):
        """
        Attaches the data to the message.

        :param message: the message to attach to
        :type message: MIMEMultipart
        :param data: the data to attach
        :type data: bytes
        :param name: the name to use for the attachment
        :type name: str
        :param mime_main: the main mime type
        :type mime_main: str
        :param mime_sub: the sub mime type
        :type mime_sub: str
        """
        # create attachment
        part = MIMEBase(mime_main, mime_sub)
        part.set_payload(data)
        encoders.encode_base64(part)

        # add header
        part.add_header(
            "Content-Disposition",
            "attachment; filename=%s" % name
        )

        # attach
        message.attach(part)
        self.logger().info("Attached: %s" % name)

    def _attach_file(self, message: MIMEMultipart, path: str):
        """
        Attaches the file to the message.

        :param message: the message to attach to
        :type message: MIMEMultipart
        :param path: the file to attach
        :type path: str
        """
        # create attachment
        with open(path, "rb") as file:
            part = MIMEBase("application", "octet-stream")
            part.set_payload(file.read())
        encoders.encode_base64(part)

        # add header
        part.add_header(
            "Content-Disposition",
            "attachment; filename=%s" % os.path.basename(path)
        )

        # attach
        message.attach(part)
        self.logger().info("Attached: %s" % path)

    @abc.abstractmethod
    def _attach_item(self, message: MIMEMultipart, item) -> bool:
        """
        Attaches the domain-specific item to the message.

        :param message: the message to attach to
        :type message: MIMEMultipart
        :param item: the item to attach
        :return: whether data type has handled
        :rtype: bool
        """
        raise NotImplementedError()

    def _attach_items(self, message: MIMEMultipart, items: List):
        """
        Attaches the items to the message.
        If an element is another list, calls itself with that list again to iterate that list.
        If an element is a string, then it gets attached as file (placeholders get automatically expanded).
        For anything else, the _attach_item method gets called.

        :param message: the message to attach to
        :type message: MIMEMultipart
        :param items: the list of items to attach
        """
        for item in items:
            if isinstance(item, list):
                self._attach_items(message, item)
            elif isinstance(item, str):
                path = self.session.expand_placeholders(item)
                self._attach_file(message, path)
            else:
                if not self._attach_item(message, item):
                    self.logger().warning("Unhandled data type: %s" % str(type(item)))

    def write_stream(self, data):
        """
        Saves the data one by one.

        :param data: the data to write (single record or iterable of records)
        """
        # initialize environment variables
        if not self._dotenv_loaded:
            self._dotenv_loaded = True
            if self.dotenv_path is None:
                path = self.session.expand_placeholders("{CWD}/.env")
                self.logger().info("Loading .env from current dir: %s" % path)
                load_dotenv(path)
            else:
                path = self.session.expand_placeholders(self.dotenv_path)
                self.logger().info("Loading .env: %s" % path)
                load_dotenv(dotenv_path=path)

        try:
            # connect
            if self._server is None:
                self._server = smtplib.SMTP(os.getenv(SMTP_HOST), int(os.getenv(SMTP_PORT)))
                if os.getenv(SMTP_STARTTLS).lower() == "true":
                    self._server.starttls()
                self.logger().info("Logging into SMTP server...")
                self._server.login(os.getenv(SMTP_USER), os.getenv(SMTP_PW))

            # assemble email
            message = MIMEMultipart()
            message["From"] = self.session.expand_placeholders(self.email_from)
            email_to = [self.session.expand_placeholders(x) for x in self.email_to]
            message["To"] = ", ".join(email_to)
            message["Subject"] = self.session.expand_placeholders(self.subject)
            body = [self.session.expand_placeholders(x) for x in self._body]
            message.attach(MIMEText("\n".join(body), "plain"))
            self._attach_items(message, make_list(data))
            self._server.send_message(message)
            self.logger().info("Email sent successfully!")
        except Exception:
            self.logger().error("Failed to send email!", exc_info=True)

    def finalize(self):
        """
        Finishes the processing, e.g., for closing files or databases.
        """
        super().finalize()
        if self._server is not None:
            try:
                self.logger().info("Logging off from SMTP server...")
                self._server.quit()
            except:
                self._server = None



import argparse
from typing import Optional, List, Dict

from wai.logging import LOGGING_WARNING
from kasperl.api import SingleVariableGenerator


class ListGenerator(SingleVariableGenerator):
    """
    Outputs the list of provided values.
    """

    def __init__(self, values: List[str] = None, var_name: str = None,
                 logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the generator.

        :param values: the starting value
        :type values: list
        :param var_name: the variable name
        :type var_name: str
        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(var_name=var_name, logger_name=logger_name, logging_level=logging_level)
        self.values = values

    def name(self) -> str:
        """
        Returns the name of the handler, used as sub-command.

        :return: the name
        :rtype: str
        """
        return "list"

    def description(self) -> str:
        """
        Returns a description of the handler.

        :return: the description
        :rtype: str
        """
        return "Outputs the specified values."

    def _default_var_name(self) -> str:
        """
        Returns the default variable name.

        :return: the default name
        :rtype: str
        """
        return "v"

    def _create_argparser(self) -> argparse.ArgumentParser:
        """
        Creates an argument parser. Derived classes need to fill in the options.

        :return: the parser
        :rtype: argparse.ArgumentParser
        """
        parser = super()._create_argparser()
        parser.add_argument("-v", "--values", type=str, metavar="VALUE", default=None, help="The list of values to use.", nargs="+")
        return parser

    def _apply_args(self, ns: argparse.Namespace):
        """
        Initializes the object with the arguments of the parsed namespace.

        :param ns: the parsed arguments
        :type ns: argparse.Namespace
        """
        super()._apply_args(ns)
        self.values = ns.values

    def _check(self) -> Optional[str]:
        """
        Hook method for performing checks.

        :return: None if checks successful, otherwise error message
        :rtype: str
        """
        result = super()._check()

        if result is None:
            if (self.values is None) or (len(self.values) == 0):
                result = "No values provided!"

        return result

    def _do_generate(self) -> List[Dict[str, str]]:
        """
        Generates the variables.

        :return: the list of variable dictionaries
        :rtype: list
        """
        result = []
        for value in self.values:
            result.append({self.var_name: str(value)})
        return result



from typing import List, Dict

from wai.logging import LOGGING_WARNING

from kasperl.api import Generator


class NullGenerator(Generator):
    """
    Outputs nothing.
    """

    def __init__(self, logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the generator.

        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(logger_name=logger_name, logging_level=logging_level)

    def name(self) -> str:
        """
        Returns the name of the handler, used as sub-command.

        :return: the name
        :rtype: str
        """
        return "null"

    def description(self) -> str:
        """
        Returns a description of the handler.

        :return: the description
        :rtype: str
        """
        return "Outputs nothing."

    def _do_generate(self) -> List[Dict[str, str]]:
        """
        Generates the variables.

        :return: the list of variable dictionaries
        :rtype: list
        """
        return []



from typing import Dict
from seppl import Plugin
from seppl.io import BatchFilter


def parse_filter(filter_: str, available_filters: Dict[str, Plugin]) -> BatchFilter:
    """
    Parses the command-line and instantiates the filter.

    :param filter_: the command-line to parse
    :type filter_: str
    :param available_filters: the available filters to use for parsing
    :type available_filters: dict
    :return: the filter instance
    :rtype: Filter
    """
    from seppl import split_args, split_cmdline, args_to_objects

    if filter_ is None:
        raise Exception("No filter command-line supplied!")
    valid = dict()
    valid.update(available_filters)
    args = split_args(split_cmdline(filter_), list(valid.keys()))
    objs = args_to_objects(args, valid, allow_global_options=False)
    if len(objs) == 1:
        if isinstance(objs[0], BatchFilter):
            result = objs[0]
        else:
            raise Exception("Expected instance of Filter but got: %s" % str(type(objs[0])))
    else:
        raise Exception("Expected to obtain one filter from '%s' but got %d instead!" % (filter_, len(objs)))
    return result



import abc
import argparse
import csv
import json
import os
from typing import List, Dict, Optional

from wai.logging import LOGGING_WARNING

from kasperl.api import make_list, SplittableStreamWriter

from seppl import AnyData, MetaDataHandler
from seppl.placeholders import placeholder_list, InputBasedPlaceholderSupporter


OUTPUT_FORMAT_TEXT = "text"
OUTPUT_FORMAT_CSV = "csv"
OUTPUT_FORMAT_JSON = "json"
OUTPUT_FORMATS = [
    OUTPUT_FORMAT_TEXT,
    OUTPUT_FORMAT_CSV,
    OUTPUT_FORMAT_JSON,
]
OUTPUT_FORMAT_EXTENSIONS = {
    OUTPUT_FORMAT_TEXT: ".txt",
    OUTPUT_FORMAT_CSV: ".csv",
    OUTPUT_FORMAT_JSON: ".json",
}


class MetaDataWriter(SplittableStreamWriter, InputBasedPlaceholderSupporter, abc.ABC):

    def __init__(self, output_dir: str = None, output_format: str = None,
                 split_names: List[str] = None, split_ratios: List[int] = None, split_group: str = None,
                 logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the reader.

        :param output_dir: the output directory to save the image/report in
        :type output_dir: str
        :param output_format: the format for the output files
        :type output_format: str
        :param split_names: the names of the splits, no splitting if None
        :type split_names: list
        :param split_ratios: the integer ratios of the splits (must sum up to 100)
        :type split_ratios: list
        :param split_group: the regular expression with a single group used for keeping items in the same split, e.g., for identifying the base name of a file or the sample ID
        :type split_group: str
        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(split_names=split_names, split_ratios=split_ratios, split_group=split_group, logger_name=logger_name, logging_level=logging_level)
        self.output_dir = output_dir
        self.output_format = output_format

    def name(self) -> str:
        """
        Returns the name of the handler, used as sub-command.

        :return: the name
        :rtype: str
        """
        return "to-metadata"

    def description(self) -> str:
        """
        Returns a description of the writer.

        :return: the description
        :rtype: str
        """
        return "Saves just the meta-data."

    def _create_argparser(self) -> argparse.ArgumentParser:
        """
        Creates an argument parser. Derived classes need to fill in the options.

        :return: the parser
        :rtype: argparse.ArgumentParser
        """
        parser = super()._create_argparser()
        parser.add_argument("-o", "--output", type=str, help="The directory to store the meta-data in. Any defined splits get added beneath there. " + placeholder_list(obj=self), required=True)
        parser.add_argument("-f", "--output_format", type=str, help="The format to use for the output, available formats: %s" % ", ".join(OUTPUT_FORMATS), required=False, default=OUTPUT_FORMAT_TEXT)
        return parser

    def _apply_args(self, ns: argparse.Namespace):
        """
        Initializes the object with the arguments of the parsed namespace.

        :param ns: the parsed arguments
        :type ns: argparse.Namespace
        """
        super()._apply_args(ns)
        self.output_dir = ns.output
        self.output_format = ns.output_format

    def accepts(self) -> List:
        """
        Returns the list of classes that are accepted.

        :return: the list of classes
        :rtype: list
        """
        return [AnyData]

    def output_text(self, meta: Dict, path: str):
        """
        Outputs the meta-data in simple textual format.

        :param meta: the meta-data
        :type meta: Dict
        :param path: the file to write to
        :type path: str
        """
        keys = sorted(meta.keys())
        with open(path, "w") as f:
            f.write("Key: Value\n")
            for k in keys:
                f.write("%s: %s\n" % (k, str(meta[k])))

    def output_csv(self, meta: Dict, path: str):
        """
        Outputs the meta-data in CSV format.

        :param meta: the meta-data
        :type meta: Dict
        :param path: the file to write to
        :type path: str
        """
        keys = sorted(meta.keys())
        with open(path, "w") as f:
            writer = csv.writer(f)
            writer.writerow(["Key", "Value"])
            for k in keys:
                writer.writerow([k, meta[k]])

    def output_json(self, meta: Dict, path: str):
        """
        Outputs the meta-data in json format.

        :param meta: the meta-data
        :type meta: Dict
        :param path: the file to write to
        :type path: str
        """
        with open(path, "w") as f:
            json.dump(meta, f, indent=2)

    @abc.abstractmethod
    def _get_name(self, item) -> Optional[str]:
        """
        Returns the name of the item.

        :param item: the item to get the name for
        :return: the name or None if not available
        :rtype: str or None
        """
        raise NotImplementedError()

    def write_stream(self, data):
        """
        Saves the data one by one.

        :param data: the data to write (single record or iterable of records)
        """
        for item in make_list(data):
            # do we have meta-data?
            if not isinstance(item, MetaDataHandler):
                self.logger().warning("Data does not handle meta-data, cannot generate output: %s" % str(item))
                continue
            # do we have a name?
            name = self._get_name(item)
            if name is None:
                self.logger().warning("Failed to determine name of item: %s" % str(item))
                continue

            sub_dir = self.session.expand_placeholders(self.output_dir)
            if self.splitter is not None:
                split = self.splitter.next(item=name)
                sub_dir = os.path.join(sub_dir, split)
            if not os.path.exists(sub_dir):
                self.logger().info("Creating dir: %s" % sub_dir)
                os.makedirs(sub_dir)

            path = os.path.join(sub_dir, name)
            path = os.path.splitext(path)[0] + OUTPUT_FORMAT_EXTENSIONS[self.output_format]
            self.logger().info("Writing meta-data to: %s" % path)
            if self.output_format == OUTPUT_FORMAT_TEXT:
                self.output_text(item.get_metadata(), path)
            elif self.output_format == OUTPUT_FORMAT_CSV:
                self.output_csv(item.get_metadata(), path)
            elif self.output_format == OUTPUT_FORMAT_JSON:
                self.output_json(item.get_metadata(), path)
            else:
                raise Exception("Unhandled output format: %s" % self.output_format)



import argparse
import csv
from typing import Optional, List, Dict

from wai.logging import LOGGING_WARNING

from kasperl.api import Generator, check_file


class CSVFileGenerator(Generator):
    """
    Forwards the values in the columns of the CSV file, using the column headers as variable names.
    """

    def __init__(self, csv_file: str = None, logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the generator.

        :param csv_file: the path to search for directories
        :type csv_file: str
        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(logger_name=logger_name, logging_level=logging_level)
        self.csv_file = csv_file

    def name(self) -> str:
        """
        Returns the name of the handler, used as sub-command.

        :return: the name
        :rtype: str
        """
        return "csv-file"

    def description(self) -> str:
        """
        Returns a description of the handler.

        :return: the description
        :rtype: str
        """
        return "Forwards the values in the columns of the CSV file, using the column headers as variable names."

    def _create_argparser(self) -> argparse.ArgumentParser:
        """
        Creates an argument parser. Derived classes need to fill in the options.

        :return: the parser
        :rtype: argparse.ArgumentParser
        """
        parser = super()._create_argparser()
        parser.add_argument("-f", "--csv_file", type=str, metavar="FILE", help="The CSV file to use.", required=True)
        return parser

    def _apply_args(self, ns: argparse.Namespace):
        """
        Initializes the object with the arguments of the parsed namespace.

        :param ns: the parsed arguments
        :type ns: argparse.Namespace
        """
        super()._apply_args(ns)
        self.csv_file = ns.csv_file

    def _check(self) -> Optional[str]:
        """
        Hook method for performing checks.

        :return: None if checks successful, otherwise error message
        :rtype: str
        """
        result = super()._check()

        if result is None:
            result = check_file(self.csv_file, "CSV", raise_exc=False)

        return result

    def _do_generate(self) -> List[Dict[str, str]]:
        """
        Generates the variables.

        :return: the list of variable dictionaries
        :rtype: list
        """
        result = []

        with open(self.csv_file, "r") as fp:
            reader = csv.DictReader(fp)
            for row in reader:
                result.append(row)

        return result



import argparse
from typing import List

from seppl import AnyData
from seppl.io import BatchFilter
from wai.logging import LOGGING_WARNING

from kasperl.api import make_list, flatten_list, NameSupporter, SourceSupporter

DUPLICATE_ACTION_IGNORE = "ignore"
DUPLICATE_ACTION_WARN = "warn"
DUPLICATE_ACTION_DROP = "drop"
DUPLICATE_ACTION_ERROR = "error"
DUPLICATE_ACTIONS = [
    DUPLICATE_ACTION_IGNORE,
    DUPLICATE_ACTION_WARN,
    DUPLICATE_ACTION_DROP,
    DUPLICATE_ACTION_ERROR,
]


class CheckDuplicateFilenames(BatchFilter):
    """
    Ensures that file names are unique (raises an exception if not).
    """

    def __init__(self, action: str = DUPLICATE_ACTION_ERROR,
                 logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the filter.

        :param action: the action to perform when encountering a duplicate file name
        :type action: str
        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(logger_name=logger_name, logging_level=logging_level)
        self.action = action
        self._names = None
        self._paths = None

    def name(self) -> str:
        """
        Returns the name of the handler, used as sub-command.

        :return: the name
        :rtype: str
        """
        return "check-duplicate-filenames"

    def description(self) -> str:
        """
        Returns a description of the handler.

        :return: the description
        :rtype: str
        """
        return "Ensures that file names are unique (raises an exception if not)."

    def accepts(self) -> List:
        """
        Returns the list of classes that are accepted.

        :return: the list of classes
        :rtype: list
        """
        return [AnyData]

    def generates(self) -> List:
        """
        Returns the list of classes that get produced.

        :return: the list of classes
        :rtype: list
        """
        return [AnyData]

    def _create_argparser(self) -> argparse.ArgumentParser:
        """
        Creates an argument parser. Derived classes need to fill in the options.

        :return: the parser
        :rtype: argparse.ArgumentParser
        """
        parser = super()._create_argparser()
        parser.add_argument("-a", "--action", choices=DUPLICATE_ACTIONS, help="The action to perform when encountering a duplicate file name", required=False, default=DUPLICATE_ACTION_ERROR)
        return parser

    def _apply_args(self, ns: argparse.Namespace):
        """
        Initializes the object with the arguments of the parsed namespace.

        :param ns: the parsed arguments
        :type ns: argparse.Namespace
        """
        super()._apply_args(ns)
        self.action = ns.action

    def initialize(self):
        """
        Initializes the processing, e.g., for opening files or databases.
        """
        super().initialize()
        self._names = set()
        self._paths = dict()

    def _do_process(self, data):
        """
        Processes the data record(s).

        :param data: the record(s) to process
        :return: the potentially updated record(s)
        """
        result = []

        for item in make_list(data):
            # name
            if not isinstance(item, NameSupporter):
                self.logger().warning("Does not implement '%s', cannot retrieve name for check, not discarding!" % str(NameSupporter))
                result.append(item)
                continue
            name = item.get_name()

            # source
            source = None
            if isinstance(item, SourceSupporter):
                source = item.get_source()

            if name in self._names:
                msg = "File name already encountered: %s" % name
                if source is not None:
                    msg += "\n- full path: %s" % source
                if name in self._paths:
                    msg += "\n- previous file: %s" % self._paths[name]

                if self.action == DUPLICATE_ACTION_IGNORE:
                    self.logger().debug(msg)
                elif self.action == DUPLICATE_ACTION_WARN:
                    self.logger().warning(msg)
                elif self.action == DUPLICATE_ACTION_DROP:
                    self.logger().warning(msg)
                    self.logger().warning("Dropping file: %s" % name)
                    item = None
                elif self.action == DUPLICATE_ACTION_ERROR:
                    raise Exception(msg)
                else:
                    raise Exception("Unhandled action: %s" % self.action)

            if item is not None:
                self._names.add(name)
                if source is not None:
                    self._paths[name] = source

        return flatten_list(result)



import argparse
from typing import List, Iterable

from wai.logging import LOGGING_WARNING

from kasperl.api import Reader
from seppl import AnyData


class StorageReader(Reader):

    def __init__(self, storage_name: str = None, logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the reader.

        :param storage_name: the name of the storage item to retrieve
        :type storage_name: str
        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(logger_name=logger_name, logging_level=logging_level)
        self.storage_name = storage_name

    def name(self) -> str:
        """
        Returns the name of the handler, used as sub-command.

        :return: the name
        :rtype: str
        """
        return "from-storage"

    def description(self) -> str:
        """
        Returns a description of the reader.

        :return: the description
        :rtype: str
        """
        return "Retrieves the specified object from internal storage."

    def _create_argparser(self) -> argparse.ArgumentParser:
        """
        Creates an argument parser. Derived classes need to fill in the options.

        :return: the parser
        :rtype: argparse.ArgumentParser
        """
        parser = super()._create_argparser()
        parser.add_argument("-s", "--storage_name", type=str, help="The name of the object to retrieve from internal storage.", required=True)
        return parser

    def _apply_args(self, ns: argparse.Namespace):
        """
        Initializes the object with the arguments of the parsed namespace.

        :param ns: the parsed arguments
        :type ns: argparse.Namespace
        """
        super()._apply_args(ns)
        self.storage_name = ns.storage_name

    def generates(self) -> List:
        """
        Returns the list of classes that get produced.

        :return: the list of classes
        :rtype: list
        """
        return [AnyData]

    def initialize(self):
        """
        Initializes the processing, e.g., for opening files or databases.
        """
        super().initialize()
        if self.storage_name is None:
            raise Exception("No storage name defined!")

    def read(self) -> Iterable:
        """
        Loads the data and returns the items one by one.

        :return: the data
        :rtype: Iterable
        """
        if self.storage_name in self.session.storage:
            self.logger().info("Retrieving from storage: %s" % self.storage_name)
            yield self.session.storage[self.storage_name]
        else:
            raise Exception("Not available from storage: %s" + self.storage_name)

    def has_finished(self) -> bool:
        """
        Returns whether reading has finished.

        :return: True if finished
        :rtype: bool
        """
        return True



import re
from typing import Tuple


COMPARISON_LESSTHAN = "lt"
COMPARISON_LESSOREQUAL = "le"
COMPARISON_EQUAL = "eq"
COMPARISON_NOTEQUAL = "ne"
COMPARISON_GREATEROREQUAL = "ge"
COMPARISON_GREATERTHAN = "gt"
COMPARISON_CONTAINS = "contains"
COMPARISON_MATCHES = "matches"

COMPARISONS = [
    COMPARISON_LESSTHAN,
    COMPARISON_LESSOREQUAL,
    COMPARISON_EQUAL,
    COMPARISON_NOTEQUAL,
    COMPARISON_GREATEROREQUAL,
    COMPARISON_GREATERTHAN,
]
COMPARISON_HELP = COMPARISON_LESSTHAN + ": less than, " \
    + COMPARISON_LESSOREQUAL + ": less or equal, " \
    + COMPARISON_EQUAL + ": equal, " \
    + COMPARISON_NOTEQUAL + ": not equal, " \
    + COMPARISON_GREATERTHAN + ": greater than, " \
    + COMPARISON_GREATEROREQUAL + ": greater of equal"

COMPARISONS_EXT = COMPARISONS[:]
COMPARISONS_EXT.append(COMPARISON_CONTAINS)
COMPARISONS_EXT.append(COMPARISON_MATCHES)
COMPARISON_EXT_HELP = COMPARISON_HELP + ", " \
                      + COMPARISON_CONTAINS + ": substring match, " \
                      + COMPARISON_MATCHES + ": regexp match"


def _ensure_same_type(v1, v2) -> Tuple:
    """
    Ensures that both values are of the same type.

    :param v1: the first value
    :param v2: the second value
    :return: the tuple of the updated values
    :rtype: tuple
    """
    if isinstance(v1, float):
        v2 = float(v2)
    elif isinstance(v1, int):
        v2 = int(v2)
    elif isinstance(v1, bool):
        v2 = str(v2).lower() == 'true'
    return v1, v2


def compare_values(v1, comparison: str, v2) -> bool:
    """
    Compares the two values using the specified comparison operator.

    :param v1: the first value
    :param comparison: the comparison to use (see COMPARISONS)
    :type comparison: str
    :param v2: the second value
    :return: the result of the comparison
    :rtype: bool
    """
    if comparison in [COMPARISON_CONTAINS, COMPARISON_MATCHES]:
        v1 = str(v1)
    else:
        v1, v2 = _ensure_same_type(v1, v2)

    # compare
    if comparison == COMPARISON_LESSTHAN:
        result = v1 < v2
    elif comparison == COMPARISON_LESSOREQUAL:
        result = v1 <= v2
    elif comparison == COMPARISON_EQUAL:
        result = v1 == v2
    elif comparison == COMPARISON_NOTEQUAL:
        result = v1 != v2
    elif comparison == COMPARISON_GREATERTHAN:
        result = v1 > v2
    elif comparison == COMPARISON_GREATEROREQUAL:
        result = v1 >= v2
    elif comparison == COMPARISON_CONTAINS:
        result = v2 in v1
    elif comparison == COMPARISON_MATCHES:
        result = re.search(v2, v1) is not None
    else:
        raise Exception("Unhandled comparison: %s" % comparison)
    return result



import abc
from typing import Dict, Any
from dataclasses import dataclass


@dataclass
class Plot(abc.ABC):
    title: str = None
    metadata: Dict[str, Any] = Dict[str, Any]


@dataclass
class XYPlot(Plot):
    x_data: Any = None
    x_label: str = None
    y_data: Any = None
    y_label: str = None


@dataclass
class SequencePlot(Plot):
    data: Any = None
    label: str = None



import argparse
from typing import Optional, List, Dict, Union

from wai.logging import LOGGING_WARNING
from kasperl.api import Generator


DEFAULT_MESSAGE = "Please enter value for variable '%s': "


class PromptGenerator(Generator):
    """
    Prompts the user to enter values for the specified variables.
    """

    def __init__(self, var_names: Union[str, List[str]] = None, message: str = None,
                 logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the generator.

        :param var_names: the variable name(s) to prompt the user with
        :type var_names: str or list
        :param message: the prompt message to use, expects one %s in the string for the variable name
        :type message: str
        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(logger_name=logger_name, logging_level=logging_level)
        self.var_names = var_names
        self.message = message

    def name(self) -> str:
        """
        Returns the name of the handler, used as sub-command.

        :return: the name
        :rtype: str
        """
        return "prompt"

    def description(self) -> str:
        """
        Returns a description of the handler.

        :return: the description
        :rtype: str
        """
        return "Prompts the user to enter values for the specified variables."

    def _create_argparser(self) -> argparse.ArgumentParser:
        """
        Creates an argument parser. Derived classes need to fill in the options.

        :return: the parser
        :rtype: argparse.ArgumentParser
        """
        parser = super()._create_argparser()
        parser.add_argument("-n", "--var_names", type=str, metavar="NAME", default=None, help="The list of variable names to prompt the user with.", nargs="+")
        parser.add_argument("-m", "--message", type=str, metavar="PROMPT", default=None, help="The custom message to prompt the user with; expects one %%s in the template which will get expanded with the variable name.", required=False)
        return parser

    def _apply_args(self, ns: argparse.Namespace):
        """
        Initializes the object with the arguments of the parsed namespace.

        :param ns: the parsed arguments
        :type ns: argparse.Namespace
        """
        super()._apply_args(ns)
        self.var_names = ns.var_names
        self.message = ns.message

    def _check(self) -> Optional[str]:
        """
        Hook method for performing checks.

        :return: None if checks successful, otherwise error message
        :rtype: str
        """
        result = super()._check()

        if result is None:
            if (self.var_names is None) or (len(self.var_names) == 0):
                result = "No variable names specified!"

        if result is None:
            if (self.message is not None) and ("%s" not in self.message):
                result = "Prompt message requires a single %s to be present!"

        return result

    def _do_generate(self) -> List[Dict[str, str]]:
        """
        Generates the variables.

        :return: the list of variable dictionaries
        :rtype: list
        """
        # prompt template
        msg = DEFAULT_MESSAGE
        if self.message is not None:
            msg = self.message

        # prompt user
        result = []
        variables = dict()
        for var_name in self.var_names:
            value = input(msg % var_name)
            variables[var_name] = value
        result.append(variables)

        return result



import argparse
import copy
import re
from typing import List

from seppl import AnyData
from seppl.io import BatchFilter
from wai.logging import LOGGING_WARNING
from kasperl.api import make_list, flatten_list, NameSupporter, SourceSupporter

SOURCE_NAME = "name"
SOURCE_PATH = "path"
SOURCES = [
    SOURCE_NAME,
    SOURCE_PATH,
]


class MetaDataFromName(BatchFilter):
    """
    Extracts a sub-string from the name or path and stores it in the meta-data.
    """

    def __init__(self, source: str = None, regexp: str = None, metadata_key: str = None,
                 logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the filter.

        :param source: where to apply the regexp to, default is name
        :type source: str
        :param regexp: the regular expression to apply to the name (1st group gets used as label)
        :type regexp: str
        :param metadata_key: the metadata key to store the extracted substring under
        :type metadata_key: str
        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(logger_name=logger_name, logging_level=logging_level)
        self.source = source
        self.regexp = regexp
        self.metadata_key = metadata_key

    def name(self) -> str:
        """
        Returns the name of the handler, used as sub-command.

        :return: the name
        :rtype: str
        """
        return "metadata-from-name"

    def description(self) -> str:
        """
        Returns a description of the filter.

        :return: the description
        :rtype: str
        """
        return "Extracts a sub-string from the name or path and stores it in the meta-data."

    def accepts(self) -> List:
        """
        Returns the list of classes that are accepted.

        :return: the list of classes
        :rtype: list
        """
        return [AnyData]

    def generates(self) -> List:
        """
        Returns the list of classes that get produced.

        :return: the list of classes
        :rtype: list
        """
        return [AnyData]

    def _create_argparser(self) -> argparse.ArgumentParser:
        """
        Creates an argument parser. Derived classes need to fill in the options.

        :return: the parser
        :rtype: argparse.ArgumentParser
        """
        parser = super()._create_argparser()
        parser.add_argument("-s", "--source", choices=SOURCES, help="The string to apply the regular expression to for extracting the meta-data.", default=SOURCE_NAME, required=False)
        parser.add_argument("-r", "--regexp", type=str, help="The regular expression apply to the name, with the 1st group being used as the meta-data value.", default=None, required=False)
        parser.add_argument("-k", "--metadata_key", type=str, help="The key in the meta-data to store the extracted sub-string under.", default=None, required=False)
        return parser

    def _apply_args(self, ns: argparse.Namespace):
        """
        Initializes the object with the arguments of the parsed namespace.

        :param ns: the parsed arguments
        :type ns: argparse.Namespace
        """
        super()._apply_args(ns)
        self.source = ns.source
        self.regexp = ns.regexp
        self.metadata_key = ns.metadata_key

    def initialize(self):
        """
        Initializes the processing, e.g., for opening files or databases.
        """
        super().initialize()
        if self.source is None:
            self.source = SOURCE_NAME
        if self.metadata_key is None:
            raise Exception("No meta-data key provided!")

    def _update(self, data):
        """
        Updates the label of the record.

        :param data: the record to update
        """
        try:
            source = None
            # get string for regexp
            if self.source == SOURCE_NAME:
                if isinstance(data, NameSupporter):
                    source = data.get_name()
            elif self.source == SOURCE_PATH:
                if isinstance(data, SourceSupporter):
                    source = data.get_source()
            else:
                raise Exception("Unhandled source: %s" % self.source)
            if source is None:
                self.logger().warning("No %s available to apply regexp to: %s" % (self.source, str(data)))
                return data

            # apply regexp
            m = re.search(self.regexp, source)
            if m is None:
                return data
            value = m.group(1)

            # set meta-data
            result = copy.deepcopy(data)
            if not result.has_metadata():
                meta = dict()
            else:
                meta = result.get_metadata()
            meta[self.metadata_key] = value
            self.logger().info("%s = %s" % (self.metadata_key, value))
            result.set_metadata(meta)
            return result
        except:
            self.logger().exception("Failed to extract meta-data value from: %s" % str(data))
            return data

    def _do_process(self, data):
        """
        Processes the data record(s).

        :param data: the record(s) to process
        :return: the potentially updated record(s)
        """
        result = []

        for item in make_list(data):
            result.append(self._update(item))

        return flatten_list(result)



import argparse
from typing import List

from seppl import MetaDataHandler, AnyData
from seppl.io import BatchFilter, FILTER_ACTIONS, FILTER_ACTION_DISCARD, FILTER_ACTION_KEEP
from wai.logging import LOGGING_WARNING

from kasperl.api import make_list, flatten_list, compare_values, \
    COMPARISONS_EXT, COMPARISON_EQUAL, COMPARISON_CONTAINS, COMPARISON_MATCHES, COMPARISON_EXT_HELP


class MetaData(BatchFilter):
    """
    Keeps or discards data records based on meta-data values.
    """

    def __init__(self, field: str = None, action: str = FILTER_ACTION_KEEP,
                 comparison: str = COMPARISON_EQUAL, value=None,
                 logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the filter.

        :param field: the name of the meta-data field to perform the comparison on
        :type field: str
        :param action: the action to perform
        :type action: str
        :param comparison: the comparison to perform
        :type comparison: str
        :param value: the value to compare with
        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(logger_name=logger_name, logging_level=logging_level)

        if action not in FILTER_ACTIONS:
            raise Exception("Invalid action: %s" % action)
        if comparison not in COMPARISONS_EXT:
            raise Exception("Invalid comparison: %s" % comparison)

        self.field = field
        self.value = value
        self.comparison = comparison
        self.action = action
        self.kept = 0
        self.discarded = 0

    def name(self) -> str:
        """
        Returns the name of the handler, used as sub-command.

        :return: the name
        :rtype: str
        """
        return "metadata"

    def description(self) -> str:
        """
        Returns a description of the filter.

        :return: the description
        :rtype: str
        """
        return "Keeps or discards data records based on meta-data comparisons. " \
               + "Performs the following comparison: METADATA_VALUE COMPARISON VALUE. " \
               + "Records that do not have the required field get discarded automatically."

    def accepts(self) -> List:
        """
        Returns the list of classes that are accepted.

        :return: the list of classes
        :rtype: list
        """
        return [AnyData]

    def generates(self) -> List:
        """
        Returns the list of classes that get produced.

        :return: the list of classes
        :rtype: list
        """
        return [AnyData]

    def _create_argparser(self) -> argparse.ArgumentParser:
        """
        Creates an argument parser. Derived classes need to fill in the options.

        :return: the parser
        :rtype: argparse.ArgumentParser
        """
        parser = super()._create_argparser()
        parser.add_argument("-f", "--field", type=str, help="The meta-data field to use in the comparison", required=True)
        parser.add_argument("-c", "--comparison", choices=COMPARISONS_EXT, default=COMPARISON_EQUAL, help="How to compare the value with the meta-data value; " + COMPARISON_EXT_HELP
                            + "; in case of '" + COMPARISON_CONTAINS + "' and '" + COMPARISON_MATCHES + "' the supplied value represents the substring to find/regexp to search with")
        parser.add_argument("-v", "--value", type=str, help="The value to use in the comparison", required=True)
        parser.add_argument("-a", "--action", choices=FILTER_ACTIONS, default=FILTER_ACTION_KEEP, help="How to react when a keyword is encountered")
        return parser

    def _apply_args(self, ns: argparse.Namespace):
        """
        Initializes the object with the arguments of the parsed namespace.

        :param ns: the parsed arguments
        :type ns: argparse.Namespace
        """
        super()._apply_args(ns)
        self.field = ns.field
        self.value = ns.value
        self.comparison = ns.comparison
        self.action = ns.action

    def initialize(self):
        """
        Initializes the processing, e.g., for opening files or databases.
        """
        super().initialize()
        if self.field is None:
            raise Exception("No meta-data field provided!")
        if self.value is None:
            raise Exception("No value provided to compare with!")
        self.kept = 0
        self.discarded = 0

    def _do_process(self, data):
        """
        Processes the data record(s).

        :param data: the record(s) to process
        :return: the potentially updated record(s)
        """
        result = []

        for item in make_list(data):
            result_item = item
            if isinstance(item, MetaDataHandler):
                meta = None
                if item.has_metadata():
                    meta = item.get_metadata()
            else:
                raise Exception("Unhandled type of data: %s" % str(type(item)))

            # no meta-data -> reject
            if meta is None:
                self.logger().info("No meta-data, discarded")
                self.discarded += 1
                continue

            # key not present -> reject
            if self.field not in meta:
                self.logger().info("Field '%s' not meta-data, discarded" % self.field)
                self.discarded += 1
                continue

            v1 = meta[self.field]
            v2 = self.value
            comp_result = compare_values(v1, self.comparison, v2)

            if self.action == FILTER_ACTION_KEEP:
                if not comp_result:
                    result_item = None
            elif self.action == FILTER_ACTION_DISCARD:
                if comp_result:
                    result_item = None
            else:
                raise Exception("Unhandled action: %s" % self.action)

            if result_item is None:
                self.discarded += 1
            else:
                self.kept += 1
                result.append(result_item)

            info = "keeping" if (result_item is not None) else "discarding"
            comp = str(meta[self.field]) + " " + self.comparison + " " + str(self.value) + " = " + str(comp_result)
            self.logger().debug("Field '%s': '%s' --> %s" % (self.field, comp, info))

        return flatten_list(result)

    def finalize(self):
        """
        Finishes the processing, e.g., for closing files or databases.
        """
        super().finalize()
        self.logger().info("# kept: %d" % self.kept)
        self.logger().info("# discarded: %d" % self.discarded)



import copy
import importlib
import inspect
import os
from typing import Optional, Union, List, Callable, Any


def strip_suffix(path: str, suffix: str) -> str:
    """
    Removes the suffix from the file, if possible.

    :param path: the filename to process
    :type path: str
    :param suffix: the suffix to remove (including extension); ignored if None or ""
    :type suffix: str
    :return: the (potentially) updated filename
    :rtype: str
    """
    if suffix is not None:
        if len(suffix) == 0:
            suffix = None
    if suffix is not None:
        if path.endswith(suffix):
            return path[0:-len(suffix)]
    return path


def locate_file(path: str, ext: Union[str, List[str]], rel_path: str = None, suffix: str = None) -> List[str]:
    """
    Tries to locate the associate files for the given path by replacing its extension by the provided ones.

    :param path: the base path to use
    :type path: str
    :param ext: the extension(s) to look for (incl dot)
    :type ext: str or list
    :param suffix: the suffix to strip from the files, ignored if None or ""
    :type suffix: str
    :param rel_path: the relative path to the annotation to use for looking for associated files, ignored if None
    :type rel_path: str
    :return: the located files
    :rtype: list
    """
    result = []
    if rel_path is not None:
        parent_path = os.path.dirname(path)
        name = os.path.basename(path)
        path = os.path.join(parent_path, rel_path, name)
    path = strip_suffix(path, suffix)
    noext = os.path.splitext(path)[0]
    for current in ext:
        path = noext + current
        if os.path.exists(path):
            result.append(path)
    return result


def load_function(function: str) -> Callable:
    """
    Parses the function definition and returns the function.
    The default format is "module_name:function_name".
    Raises exceptions if wrong format, missing or not an actual function.

    :param function: the function definition to parse
    :type function: str
    :return: the parsed function
    """
    if ":" not in function:
        raise Exception("Expected format 'module_name:function_name' but got: %s" % function)
    else:
        module_name, func_name = function.split(":")

    try:
        module = importlib.import_module(module_name)
    except:
        raise Exception("Failed to import class lister module: %s" % module_name)

    if hasattr(module, func_name):
        func = getattr(module, func_name)
        if inspect.isfunction(func):
            return func
        else:
            raise Exception("Not an actual function: %s" % function)
    else:
        raise Exception("Function '%s' not found in module '%s'!" % (func_name, module_name))


def safe_deepcopy(obj: Optional[Any]) -> Optional[Any]:
    """
    Creates a deep copy of the object. Skips None objects.

    :param obj: the object to copy, can be None
    :return: the copy or None
    """
    if obj is None:
        return None
    else:
        return copy.deepcopy(obj)


def annotation_to_name(ann: str, ext: str = ".jpg") -> str:
    """
    Generates a name from the annotation file name.

    :param ann: the annotation to create the name from
    :type ann: str
    :param ext: the extension to use for the name (incl dot)
    :type ext: str
    :return: the generated name
    :rtype: str
    """
    result = os.path.basename(ann)
    result = os.path.splitext(result)[0] + ext
    return result


def check_dir(path: str, name: str, raise_exc: bool = True) -> Optional[str]:
    """
    Performs checks on the specified directory.
    Raises an exception if a problem is encountered (not None, must exist, must be a directory).

    :param path: the directory to check
    :type path: str
    :param name: the name of the directory
    :type name: str
    :param raise_exc: whether to raise an exception rather than returning the error message
    :type raise_exc: bool
    :return: the error messages or None if everything fine
    :rtype: str
    """
    result = None
    if path is None:
        result = "No %s directory provided!" % name
    elif not os.path.exists(path):
        result = "%s directory does not exist: %s" % (name, path)
    elif not os.path.isdir(path):
        result = "%s directory does not point to a directory: %s" % (name, path)
    if raise_exc and (result is not None):
        raise Exception(result)
    return result


def check_file(path: str, name: str, raise_exc: bool = True) -> Optional[str]:
    """
    Performs checks on the specified file.
    Raises an exception if a problem is encountered (not None, must exist, must not be a directory).

    :param path: the file to check
    :type path: str
    :param name: the name of the file
    :type name: str
    :param raise_exc: whether to raise an exception rather than returning the error message
    :type raise_exc: bool
    :return: the error messages or None if everything fine
    :rtype: str
    """
    result = None
    if path is None:
        result = "No %s file provided!" % name
    elif not os.path.exists(path):
        result = "%s file does not exist: %s" % (name, path)
    elif os.path.isdir(path):
        result = "%s points to a directory: %s" % (name, path)
    if raise_exc and (result is not None):
        raise Exception(result)
    return result



from ._csv_file import CSVFileGenerator
from ._dirs import DirectoryGenerator, VARS, VAR_ABSDIR, VAR_RELDIR, VAR_DIRNAME
from ._list import ListGenerator
from ._null import NullGenerator
from ._prompt import PromptGenerator
from ._range import RangeGenerator
from ._text_file import TextFileGenerator



import argparse
import os
from typing import List, Iterable

from wai.logging import LOGGING_WARNING

from kasperl.api import Reader
from seppl.placeholders import placeholder_list, PlaceholderSupporter


class TextFileReader(Reader, PlaceholderSupporter):

    def __init__(self, path: str = None, logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the reader.

        :param path: the path of the text file to load
        :type path: str
        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(logger_name=logger_name, logging_level=logging_level)
        self.path = path
        self._finished = False

    def name(self) -> str:
        """
        Returns the name of the handler, used as sub-command.

        :return: the name
        :rtype: str
        """
        return "from-text-file"

    def description(self) -> str:
        """
        Returns a description of the reader.

        :return: the description
        :rtype: str
        """
        return "Reads the specified text file line by line and forwards the data."

    def _create_argparser(self) -> argparse.ArgumentParser:
        """
        Creates an argument parser. Derived classes need to fill in the options.

        :return: the parser
        :rtype: argparse.ArgumentParser
        """
        parser = super()._create_argparser()
        parser.add_argument("-p", "--path", metavar="FILE", type=str, help="The file to load; " + placeholder_list(obj=self), required=True)
        return parser

    def _apply_args(self, ns: argparse.Namespace):
        """
        Initializes the object with the arguments of the parsed namespace.

        :param ns: the parsed arguments
        :type ns: argparse.Namespace
        """
        super()._apply_args(ns)
        self.path = ns.path

    def generates(self) -> List:
        """
        Returns the list of classes that get produced.

        :return: the list of classes
        :rtype: list
        """
        return [str]

    def initialize(self):
        """
        Initializes the processing, e.g., for opening files or databases.
        """
        super().initialize()
        if self.path is None:
            raise Exception("No path specified!")

    def read(self) -> Iterable:
        """
        Loads the data and returns the items one by one.

        :return: the data
        :rtype: Iterable
        """
        path = self.session.expand_placeholders(self.path)
        if not os.path.exists(path):
            raise Exception("Path does not exist: %s" % path)
        if not os.path.isfile(path):
            raise Exception("Not a file: %s" % path)
        self.logger().info("Reading from: %s" % path)
        self._finished = False
        with open(path, "r") as fp:
            for line in fp.readlines():
                yield line.strip()
        self._finished = True

    def has_finished(self) -> bool:
        """
        Returns whether reading has finished.

        :return: True if finished
        :rtype: bool
        """
        return self._finished



import argparse
from dataclasses import dataclass
from typing import List, Any, Optional


OPTION_LIST_MAX = 72
DESCRIPTION_OFFSET = 22


@dataclass
class CommandlineParameter:
    """
    Container for defining options for the argument parser.
    """
    short_opt: str = None
    long_opt: str = None
    metavar: str = None
    choices: List[str] = None
    help: str = None
    type: Any = None
    action: str = None
    required: bool = False
    default: Any = None
    is_help: bool = False  # whether a help parameter, gets skipped when building the parser
    nargs: str = None


def param_to_short(param: CommandlineParameter) -> str:
    """
    Turns a parameter definition into a short string for the command list.

    :param param: the parameter to convert into a string
    :type param: CommandlineParameter
    :return: the generated string
    :rtype: str
    """
    if param.short_opt is not None:
        result = param.short_opt
    else:
        result = param.long_opt

    if param.choices is not None:
        result += " {" + ",".join(param.choices) + "}"
    elif param.metavar is not None:
        result += " " + param.metavar

    if not param.required:
        result = "[" + result + "]"

    return result


def params_to_short(prog: str, params: List[CommandlineParameter], additional: str = None) -> str:
    """
    Turns the executable and the parameters into a short overview string.

    :param prog: the executable
    :type prog: str
    :param params: the list of parameters to process
    :type params: list
    :param additional: additional strings to add, ignored if None or empty
    :type additional: str
    :return: the generated string
    :rtype: str
    """
    result = []
    line = "usage: " + prog
    prefix = " " * len(line)
    for param in params:
        param_short = param_to_short(param)
        if len(line) + 1 + len(param_short) > OPTION_LIST_MAX:
            result.append(line)
            line = prefix
        line += " " + param_short
    if len(line) > 0:
        result.append(line)

    if (additional is not None) and (len(additional) > 0):
        result.append(prefix + additional)

    return "\n".join(result)


def param_to_help(param: CommandlineParameter) -> str:
    """
    Generates a long help description for the parameter.

    :param param: the parameter to turn into a help description
    :type param: CommandlineParameter
    :return: the generate help description
    :rtype: str
    """
    # assemble choices
    if param.choices is not None:
        choices = " {" + ",".join(param.choices) + "}"
    else:
        choices = ""

    result = "  "
    if (param.short_opt is None) and (param.long_opt is None):
        raise Exception("Either short or long option flag needs to be provided: %s" % str(param))
    elif (param.short_opt is not None) and (param.long_opt is not None):
        result += param.short_opt + choices + ", " + param.long_opt + choices
    elif param.short_opt is not None:
        result += param.short_opt + choices
    else:
        result += param.long_opt + choices
    if param.metavar is not None:
        result += " " + param.metavar

    # align help text
    if len(result) <= DESCRIPTION_OFFSET:
        result += " " * DESCRIPTION_OFFSET
        result = result[0:DESCRIPTION_OFFSET]
        result += " " + param.help
    else:
        result += "\n" + " " * DESCRIPTION_OFFSET + " " + param.help

    return result


def param_to_parser(parser: argparse.ArgumentParser, param: CommandlineParameter):
    """
    Adds the parameter to the argument parser.
    Skips parameters that are flagged with "is_help=True", as they are handled separately.

    :param parser: the parser to append
    :type parser: argparse.ArgumentParser
    :param param: the parameter to add
    :type param: CommandlineParameter
    """
    if param.is_help:
        return

    args = []
    if param.short_opt is not None:
        args.append(param.short_opt)
    if param.long_opt is not None:
        args.append(param.long_opt)

    kwargs = {}
    if param.metavar is not None:
        kwargs["metavar"] = param.metavar
    if param.choices is not None:
        kwargs["choices"] = param.choices
    if param.help is not None:
        kwargs["help"] = param.help
    if param.action is not None:
        kwargs["action"] = param.action
    if param.type is not None:
        kwargs["type"] = param.type
    if param.nargs is not None:
        kwargs["nargs"] = param.nargs
    kwargs["required"] = param.required
    kwargs["default"] = param.default

    parser.add_argument(*args, **kwargs)


def params_to_parser(parser: argparse.ArgumentParser, params: Optional[List[CommandlineParameter]]):
    """
    Adds all the parameters to the parser.
    Skips parameters that are flagged with "is_help=True", as they are handled separately.

    :param parser: the parser to append
    :type parser: argparse.ArgumentParser
    :param params: the parameters to add
    :type params: list
    """
    if params is None:
        return
    for param in params:
        param_to_parser(parser, param)



import argparse
import logging
import os
import shlex
from typing import List, Dict, Optional, Union

from wai.logging import init_logging, set_logging_level, add_logging_level

from seppl import split_cmdline, Plugin, load_args
from seppl.placeholders import load_user_defined_placeholders, expand_placeholders
from ._generator import compile_generator_vars_list
from ._help import CommandlineParameter, params_to_parser


PARAM_EXEC_FORMAT = "--exec_format"


DESCRIPTION = "Tool for executing a pipeline multiple times, each time with a different set of variables expanded. " \
              + "A variable is surrounded by curly quotes (e.g., variable 'i' gets referenced with '{i}'). " \
              + "When supplying multiple generators, then these get treated as nested executions."


PIPELINE_FORMAT_CMDLINE = "cmdline"
PIPELINE_FORMAT_FILE = "file"
PIPELINE_FORMATS = [
    PIPELINE_FORMAT_CMDLINE,
    PIPELINE_FORMAT_FILE,
]


def load_pipeline(pipeline: Union[str, List[str]], pipeline_format: str = PIPELINE_FORMAT_CMDLINE,
                  remove_convert_prog: bool = False, convert_prog: str = None, logger: logging.Logger = None) -> List[str]:
    """
    Loads the pipeline according to the format and returns the split arguments.

    :param pipeline: the pipeline
    :param pipeline_format: the format of the pipeline
    :type pipeline_format: str
    :param remove_convert_prog: whether to remove the convert program from the arguments
    :type remove_convert_prog: bool
    :param convert_prog: the executable to potentially remove
    :type convert_prog: str
    :param logger: the optional logger to use
    :type logger: logging.Logger
    :return: the split arguments of the pipeline
    :rtype: list
    """
    if pipeline_format == PIPELINE_FORMAT_CMDLINE:
        if isinstance(pipeline, str):
            result = split_cmdline(pipeline)
        else:
            result = pipeline

    elif pipeline_format == PIPELINE_FORMAT_FILE:
        if isinstance(pipeline, str):
            pipeline_file = pipeline
        else:
            pipeline_file = shlex.join(pipeline)
        pipeline_file = expand_placeholders(pipeline_file)
        result = load_args(pipeline_file, logger=logger)

    else:
        raise Exception("Unhandled pipeline format: %s" % pipeline_format)

    # remove convert prog from pipeline
    if remove_convert_prog:
        if len(result) > 0:
            if result[0].startswith(convert_prog):
                result = result[1:]

    if logger is not None:
        logger.info("Pipeline: %s" % str(result))

    return result


def _expand_vars(pipeline: List[str], vars_: Dict) -> Union[str, List[str]]:
    """
    Expands the variables in the pipeline.

    :param pipeline: the pipeline to expand
    :type pipeline: list
    :param vars_: the variables to expand
    :type vars_: dict
    :return: the expanded pipeline
    :rtype: str or list
    """
    result = []
    for arg in pipeline:
        for var in vars_:
            arg = arg.replace("{%s}" % var, vars_[var])
        result.append(arg)

    return result


def execute_pipeline(convert_prog: str, convert_func, pipeline: Union[str, List[str]],
                     generator: Union[str, List[str]], generator_plugins: Dict[str, Plugin],
                     pipeline_format: str = PIPELINE_FORMAT_CMDLINE, dry_run: bool = False,
                     prefix: str = False, logger: logging.Logger = None):
    """
    Executes the specified pipeline as many times as the generators produce variables.

    :param convert_prog: the conversion executable
    :type convert_prog: str
    :param convert_func: the main method for executing the conversion executable
    :param pipeline: the pipeline template to use
    :type pipeline: str or list
    :param generator: the generator command-line(s) to use for generating variable values to be expanded in the pipeline template
    :type generator: str or list
    :param generator_plugins: the available generator plugins (name -> plugin)
    :type generator_plugins: dict
    :param pipeline_format: the format the pipeline is in
    :type pipeline_format: str
    :param dry_run: whether to only expand/output but not execute the pipeline
    :type dry_run: bool
    :param prefix: the prefix to use when in dry-run mode
    :type prefix: str
    :param logger: the logger instance to use
    :type logger: logging.Logger
    """
    # load pipeline
    pipeline = load_pipeline(pipeline, pipeline_format=pipeline_format,
                             remove_convert_prog=True, convert_prog=convert_prog, logger=logger)

    # generate variables
    vars_list = compile_generator_vars_list(generator, generator_plugins)

    # apply generator to pipeline template and execute it
    for vars_ in vars_list:
        pipeline_exp = _expand_vars(pipeline, vars_)
        if logger is not None:
            logger.info("%s\n--> %s" % (str(pipeline), str(pipeline_exp)))
        if dry_run:
            if prefix is not None:
                if not prefix.endswith(" "):
                    prefix = prefix + " "
                pipeline_exp[0] = prefix + pipeline_exp[0]
                print(shlex.join(pipeline_exp))
        else:
            convert_func(pipeline_exp)


def perform_pipeline_execution(env_var: Optional[str], args: List[str], prog: str, description: Optional[str],
                               convert_prog: str, convert_func, generators_plugins: Dict[str, Plugin],
                               logger: logging.Logger, additional_params: Optional[List[CommandlineParameter]] = None,
                               pre_exec=None, post_exec=None):
    """
    The main method for parsing command-line arguments.

    :param env_var: the environment variable to obtain the logging level from, can be None
    :type env_var: str
    :param args: the commandline arguments, uses sys.argv if not supplied
    :type args: list
    :param prog: the executable name
    :type prog: str
    :param description: the description for the executable, uses default if None; automatically appends available generators
    :type description: str
    :param convert_prog: the conversion executable
    :type convert_prog: str
    :param convert_func: the main method for executing the conversion executable
    :param generators_plugins: the available generators
    :type generators_plugins: dict
    :param logger: the logger instance to use
    :type logger: logging.Logger
    :param additional_params: the additional parameters for the parser
    :type additional_params: list
    :param pre_exec: the optional method to execute before the pipeline executions, gets parsed args namespace as only argument
    :param post_exec: the optional method to execute after the pipeline got executed, gets parsed args namespace as only argument
    """
    init_logging(env_var=env_var)

    if description is None:
        description = DESCRIPTION
    description += " Available generators: " + ", ".join(generators_plugins)
    parser = argparse.ArgumentParser(
        prog=prog, description=description,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument("--exec_generator", metavar="GENERATOR", help="The generator plugin(s) to use, incl. their options. Flag needs to be specified for each generator.", default=None, type=str, required=True, action="append")
    parser.add_argument("--exec_dry_run", action="store_true", help="Applies the generator to the pipeline template and only outputs it on stdout.", required=False)
    parser.add_argument("--exec_prefix", metavar="PREFIX", help="The string to prefix the pipeline with when in dry-run mode.", required=False, default=None, type=str)
    parser.add_argument("--exec_placeholders", metavar="FILE", help="The file with custom placeholders to load (format: key=value).", required=False, default=None, type=str)
    parser.add_argument(PARAM_EXEC_FORMAT, choices=PIPELINE_FORMATS, required=False, default=PIPELINE_FORMAT_CMDLINE,
                        help="The format that the pipeline is in. "
                             + "The format '" + PIPELINE_FORMAT_CMDLINE + "' interprets the remaining "
                             + "arguments as the pipeline arguments to execute. "
                             + "The format '" + PIPELINE_FORMAT_FILE + "' expects a file to load the pipeline arguments "
                             + "from. This file format allows spreading the pipeline arguments over multiple lines: it "
                             + "simply joins all lines into a single command-line before splitting it into individual "
                             + "arguments for execution.")
    parser.add_argument("pipeline", help="The pipeline template with variables to expand and then execute; see '" + PARAM_EXEC_FORMAT + "' option.", nargs=argparse.REMAINDER)
    if logger is not None:
        add_logging_level(parser, short_opt=None, long_opt="--exec_logging_level")
    params_to_parser(parser, additional_params)

    parsed = parser.parse_args(args=args)

    if logger is not None:
        set_logging_level(logger, parsed.logging_level)

    if parsed.exec_placeholders is not None:
        if not os.path.exists(parsed.exec_placeholders):
            msg = "Placeholder file not found: %s" % parsed.exec_placeholders
            if logger is not None:
                logger.error(msg)
            else:
                print(msg)
        else:
            msg = "Loading custom placeholders from: %s" % parsed.exec_placeholders
            if logger is not None:
                logger.info(msg)
            else:
                print(msg)
            load_user_defined_placeholders(parsed.exec_placeholders)

    # custom pre-execution hook?
    if pre_exec is not None:
        pre_exec(parsed)

    execute_pipeline(convert_prog, convert_func, parsed.pipeline, parsed.exec_generator, generators_plugins,
                     pipeline_format=parsed.exec_format, dry_run=parsed.exec_dry_run, prefix=parsed.exec_prefix,
                     logger=logger)

    # custom post-execution hook?
    if post_exec is not None:
        post_exec(parsed)



import argparse
import os
import re
import traceback
from typing import Optional, List, Dict, Tuple

from wai.logging import LOGGING_WARNING
from kasperl.api import Generator


VAR_ABSDIR = "absdir"
VAR_RELDIR = "reldir"
VAR_DIRNAME = "dirname"
VARS = [
    VAR_ABSDIR,
    VAR_RELDIR,
    VAR_DIRNAME,
]


class DirectoryGenerator(Generator):
    """
    Iterates over directories that it finds.
    """

    def __init__(self, path: str = None, recursive: bool = False, regexp: str = None, file_regexp: str = None,
                 logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the generator.

        :param path: the path to search for directories
        :type path: str
        :param recursive: whether to search recursively
        :type recursive: bool
        :param regexp: the regular expression for matching directories
        :type regexp: str
        :param file_regexp: the regular expression that at least one file must match in a directory (path is excluded from test), ignored if None
        :param file_regexp: str
        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(logger_name=logger_name, logging_level=logging_level)
        self.path = path
        self.recursive = recursive
        self.regexp = regexp
        self.file_regexp = file_regexp

    def name(self) -> str:
        """
        Returns the name of the handler, used as sub-command.

        :return: the name
        :rtype: str
        """
        return "dirs"

    def description(self) -> str:
        """
        Returns a description of the handler.

        :return: the description
        :rtype: str
        """
        return "Iterates over directories that it finds. Can be limited to directories that contain certain files. " \
            + "Available variables: " + "|".join(VARS) + ". " \
            + VAR_ABSDIR + ": the absolute directory, " \
            + VAR_RELDIR + ": the relative directory to the search path, " \
            + VAR_DIRNAME + ": the directory name (no parent path)."

    def _create_argparser(self) -> argparse.ArgumentParser:
        """
        Creates an argument parser. Derived classes need to fill in the options.

        :return: the parser
        :rtype: argparse.ArgumentParser
        """
        parser = super()._create_argparser()
        parser.add_argument("-p", "--path", type=str, metavar="DIR", default=None, help="The directory/directories to search", required=True, nargs="+")
        parser.add_argument("-r", "--recursive", action="store_true", help="Whether to search for directories recursively.", required=False)
        parser.add_argument("--regexp", type=str, metavar="REGEXP", default=None, help="The regular expression to use for matching directories; matches all if not provided.", required=False)
        parser.add_argument("--file_regexp", type=str, metavar="REGEXP", default=None, help="Only directories that have at least one file matching this regexp are returned (path is excluded from test); all directories are turned if not provided.", required=False)
        return parser

    def _apply_args(self, ns: argparse.Namespace):
        """
        Initializes the object with the arguments of the parsed namespace.

        :param ns: the parsed arguments
        :type ns: argparse.Namespace
        """
        super()._apply_args(ns)
        self.path = ns.path
        self.recursive = ns.recursive
        self.regexp = ns.regexp
        self.file_regexp = ns.file_regexp

    def _check(self) -> Optional[str]:
        """
        Hook method for performing checks.

        :return: None if checks successful, otherwise error message
        :rtype: str
        """
        result = super()._check()

        if result is None:
            if self.regexp == "":
                self.regexp = None
            if self.regexp is not None:
                try:
                    re.compile(self.regexp)
                except:
                    result = "Invalid regular expression: %s\n%s" % (self.regexp, traceback.format_exc())

        if result is None:
            if self.file_regexp is not None:
                try:
                    re.compile(self.file_regexp)
                except:
                    result = "Invalid regular expression for files: %s\n%s" % (self.file_regexp, traceback.format_exc())

        if result is None:
            for p in self.path:
                if not os.path.exists(p):
                    result = "Directory does not exist: %s" % p
                elif not os.path.isdir(p):
                    result = "Not a directory: %s" % p

        return result

    def _has_file_matches(self, path: str) -> bool:
        """
        Checks whether the path has any matching files (if the regexp for files is specified).

        :param path: the path to check
        :type path: str
        :return: True if matching files or no regexp for files
        :rtype: bool
        """
        if self.file_regexp is None:
            return True
        for f in os.listdir(path):
            m = re.match(self.file_regexp, f)
            if m is not None:
                return True
        return False

    def _locate(self, start: str, current: str, recursive: bool, paths: List[Tuple[str, str]]):
        """
        Locates directories.

        :param start: the starting directory (for determining relative dirs)
        :type start: str
        :param current: the directory to search
        :type current: str
        :param recursive: whether to search recursively
        :type recursive: bool
        :param paths: for collecting the matching dirs
        :type paths: list
        """
        for f in os.listdir(current):
            full = os.path.join(current, f)
            if os.path.isdir(full):
                if self.regexp is not None:
                    m = re.match(self.regexp, f)
                    if (m is not None) and self._has_file_matches(full):
                        paths.append((start, full))
                else:
                    if self._has_file_matches(full):
                        paths.append((start, full))
                if recursive:
                    self._locate(start, full, recursive, paths)

    def _do_generate(self) -> List[Dict[str, str]]:
        """
        Generates the variables.

        :return: the list of variable dictionaries
        :rtype: list
        """
        result = []

        # locate dirs
        paths = []
        for abs_dir in self.path:
            self._locate(os.path.abspath(abs_dir), os.path.abspath(abs_dir), self.recursive, paths)

        # prepare variables
        for parent_dir, abs_dir in paths:
            if abs_dir.startswith(parent_dir):
                rel_dir = abs_dir[len(parent_dir):]
                if rel_dir.startswith("/") or rel_dir.startswith("\\"):
                    rel_dir = rel_dir[1:]
            else:
                rel_dir = None
            dir_name = os.path.basename(abs_dir)
            result.append({
                VAR_ABSDIR: abs_dir,
                VAR_RELDIR: rel_dir,
                VAR_DIRNAME: dir_name,
            })

        return result



import argparse
from random import Random
from typing import List

from seppl import AnyData
from seppl.io import BatchFilter
from wai.logging import LOGGING_WARNING
from kasperl.api import flatten_list, make_list


class Sample(BatchFilter):
    """
    Selects a sub-sample from the stream.
    """

    def __init__(self, seed: int = None, threshold: float = 0.0,
                 logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the filter.

        :param seed: the seed value to use for the random number generator; randomly seeded if not provided
        :type seed: int
        :param threshold: the threshold to use for Random.rand(): if equal or above, sample gets selected; range: 0-1; default: 0 (= always)
        :type threshold: float
        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(logger_name=logger_name, logging_level=logging_level)
        self.seed = seed
        self.threshold = threshold
        self._random = None

    def name(self) -> str:
        """
        Returns the name of the handler, used as sub-command.

        :return: the name
        :rtype: str
        """
        return "sample"

    def description(self) -> str:
        """
        Returns a description of the filter.

        :return: the description
        :rtype: str
        """
        return "Selects a sub-sample from the stream."

    def accepts(self) -> List:
        """
        Returns the list of classes that are accepted.

        :return: the list of classes
        :rtype: list
        """
        return [AnyData]

    def generates(self) -> List:
        """
        Returns the list of classes that get produced.

        :return: the list of classes
        :rtype: list
        """
        return [AnyData]

    def _create_argparser(self) -> argparse.ArgumentParser:
        """
        Creates an argument parser. Derived classes need to fill in the options.

        :return: the parser
        :rtype: argparse.ArgumentParser
        """
        parser = super()._create_argparser()
        parser.add_argument("-s", "--seed", type=int, help="The seed value to use for the random number generator; randomly seeded if not provided.", default=None, required=False)
        parser.add_argument("-T", "--threshold", type=float, help="The threshold to use for Random.rand(): if equal or above, sample gets selected; range: 0-1; 0 = always.", default=0.0, required=False)
        return parser

    def _apply_args(self, ns: argparse.Namespace):
        """
        Initializes the object with the arguments of the parsed namespace.

        :param ns: the parsed arguments
        :type ns: argparse.Namespace
        """
        super()._apply_args(ns)
        self.seed = ns.seed
        self.threshold = ns.threshold

    def initialize(self):
        """
        Initializes the processing, e.g., for opening files or databases.
        """
        super().initialize()
        self._random = Random(self.seed)
        threshold = 0.0 if self.threshold is None else self.threshold
        if (threshold < 0) or (threshold > 1):
            raise Exception("Threshold must satisfy x >= 0 and x <= 1, supplied: %f" % threshold)

    def _do_process(self, data):
        """
        Processes the data record(s).

        :param data: the record(s) to process
        :return: the potentially updated record(s)
        """
        result = []

        for item in make_list(data):
            threshold = 0.0 if self.threshold is None else self.threshold
            if self._random.random() >= threshold:
                result.append(item)

        return flatten_list(result)
